# frozen_string_literal: true

User.class_eval do
  field :main_applicant_id, type: BSON::ObjectId
  field :reservation_holder_id, type: BSON::ObjectId
  field :lessor_id, type: BSON::ObjectId
  field :gender, type: String
  field :id_cristal, type: String

  index({main_applicant_id: 1}, {background: true})
  index({reservation_holder_id: 1}, {background: true})
  index({lessor_id: 1}, {background: true})
  index({gender: 1}, {background: true})
  index({id_cristal: 1}, {background: true})
  index({updated_at: 1}, {background: true})
  index({created_at: 1}, {background: true})
  index({current_sign_in_at: 1}, {background: true})


  before_save :hermes_before_actions
  after_save :hermes_after_actions
  after_create :hermes_after_create_actions
  after_destroy :destroy_user_rights 


  has_many(
    :notifications,
    class_name: '::NotificationCenter::Notification',
    foreign_key: :targetable_id,
    dependent: :destroy,
    order: 'created_at DESC'
  )

  ###############################
  #### Notification #############
  include ::Concerns::NotificationCenter::Notifiable

  def self.targets_selector(hermes_user, opt = {})
    [hermes_user].compact
  end

  def self.find_first_by_auth_conditions(warden_conditions)
    conditions = warden_conditions.dup
    if username = conditions.delete(:username).downcase
      self.any_of({ :login =>  username }, { :email =>  username }).first ||
      self.any_of({ :login =>  /^#{::Regexp.escape(username)}$/i }, { :email =>  /^#{::Regexp.escape(username)}$/i }).first
    else
      super
    end
  end

  def self.decode_jwt_token(token)
    JWT.decode token, APP_CONFIG["token"], true, { algorithm: 'HS256' }
  end 

  def notification_mailer
    ::Hermes::HermesMailer
  end

  def notifiable_title(key, opt={})
    I18n.t("#{key}.title")
  end

  def notifiable_body(key, opt={})
    I18n.t("#{key}.body")
  end

  def notifiable_path
    ""
  end

  ###############################
  ###############################

  def has_role?(role)
    ((roles || []) + (auth_roles || [])).include? role
  end

  # user.has_roles?('role1', 'role2') or user.has_roles?(%[role1 role2])
  def has_roles?(*args)
    if (args.count == 1) && args.first.is_a?(Array)
      return (((roles || []) + (auth_roles || [])) & args.first).any?
    end
    (((roles || []) + (auth_roles || [])) & args).any?
  end

  def main_applicant
    Hermes::Applicant.find_by(id: main_applicant_id)
  end

  def reservation_holder
    Hermes::ReservationHolder.find(reservation_holder_id)
  end

  def lessor
    Hermes::Lessor.find(lessor_id)
  rescue
    nil
  end

  def lessor=(les)
    self.lessor_id = les.id
  end

  def create_user_on_identity_provider
  end

  def get_identity_provider
    IdentityProvider.find_by(key: provider || HERMES_CONFIG['identity_provider_key']) 
  end 

  def create_new_user_on_identity_provider(already_check = 0 )
    if create_user_data && 
      !create_user_data.empty? && 
      identity_provider = get_identity_provider
      
      @create_user_response = send_create_user_request_to_identity_provider(identity_provider, create_user_data)
      update_user_roles_beiam(already_check)
      
      return false unless [200, 201, 409].include?(@create_user_response.code)
    end
  end

  def login_or_email
    login || email 
  end 
  
  def update_user_roles_beiam(already_check = 0)
    return if Rails.env.test?
    return unless identity_provider = get_identity_provider
    return unless identity_provider.add_roles_user_url
    url_get = URI.escape(identity_provider.reload.add_roles_user_url.sub!('{login}', login_or_email).sub!('{role}', ""))
    token = get_beiam_token(identity_provider)
    data = {headers: {'Authorization': "Bearer #{token}"}}
    response = HTTParty.get(url_get, data) 

    if already_check >= 2
      Hermes::LogJob.add_to_queue(level: 'error', :code => "BEIAM-LOGIN-ERROR", :action => "Impossible de mettre à jour les roles de #{email} sur beiam", user_id: user.id.to_s)
      raise "Update user auth roles on beiam failed for email #{email}"
      return
    end 

    if response.code == 404
      create_new_user_on_identity_provider(already_check + 1)
      return
    end

    roles = response.code == 200 ? JSON.parse(response.body).map { |r| r['name']} : []
    
    added_roles = (self.auth_roles || []) - (roles || [])
    deleted_roles = (roles || []) - (self.auth_roles  || []) 
    
    added_roles.each do |role|
      send_add_roles_beiam(identity_provider, role, true)
    end
    
    deleted_roles.each do |role|
      send_add_roles_beiam(identity_provider, role, false)
    end
  end
  
  def send_create_user_request_to_identity_provider(identity_provider, create_user_data = {}, update = false)
    url = identity_provider.create_user_url
    url += "/" + login_or_email if update
    
    data = {
      headers: identity_provider.create_user_headers,
      body: create_user_data.to_json,
    }
  
    update ? HTTParty.put(url, data) : HTTParty.post(url, data) 
  end
  
  def send_add_roles_beiam(identity_provider, role, add)
    token = get_beiam_token(identity_provider)
    url = URI.escape(identity_provider.reload.add_roles_user_url.sub!('{login}', login_or_email).sub!('{role}', role))
    data = {headers: {'Authorization': "Bearer #{token}"}}
    
    add ? HTTParty.put(url, data) : HTTParty.delete(url, data)
  end

  def get_beiam_token(identity_provider)
    RestClient.post(
      identity_provider.account_service_url,
      headers: { "Content-Type" => "application/x-www-form-urlencoded" },
      client_id: identity_provider.account_service_client_id,
      client_secret: identity_provider.account_service_client_secret,
      grant_type: 'client_credentials'
    ).json_to_h[:access_token]
  end

  def has_lessor?
    lessor_id.present?
  end

  def hermes_before_actions
    check_roles if auth_roles_changed?
    self.provider = HERMES_CONFIG['identity_provider_key'] if provider.nil?
  end

  def hermes_after_actions
    update_user_roles_beiam if auth_roles_changed? || email_changed? || id_cristal_changed?
  end

  def check_roles
    self.auth_roles = self.auth_roles - ['entreprise', 'bailleur'] unless (['cre', 'rdsl'] & self.auth_roles).blank?
  end 

  def hermes_after_create_actions
    Hermes::SignupUserNotificationPreference.update_user_preference(self)
    notify(key: NOTIFICATION_KEYS[:user][:create], channels: [])
  end

  def destroy_user_rights
    Hermes::UserRight.where(user_id: id).destroy_all
  end
  
  def entity_lessor_with_user_right(type ='entity')
    ids = Hermes::UserRight.where(user_id: id, type: type).pluck(:"#{type}_id")
    model = type == 'entity' ? Hermes::Entity : Hermes::Lessor
    model.where(:id.in => ids)
  end 

  field :is_slow, type: Boolean, default: false # slow notification (like email) will be sent in a dedicated queue
 
  def notify(opt={})
    key = opt.fetch(:key, 'default')
    level = opt.fetch(:level, 'all')
    sender = opt.fetch(:sender, nil)
    channels = opt.fetch(:channels, %w[email web slack sms])
    datas = opt.fetch(:datas, {})
    is_slow = opt.fetch(:is_slow, false)

    self.class.targets_selector(self, opt).each do |target|
      notifiable_job = is_slow ? SlowNotifierJob : AsyncNotifierJob
      
      notifiable_job.perform_later(
        title: notifiable_title(key, opt),
        body: notifiable_body(key, opt),
        notifiable_type: self.class.name,
        notifiable_id: self.id.to_s,
        targetable_type: target.class.name,
        targetable_id: target.id.to_s,
        sender_type: sender&.class&.name,
        sender_id: sender&.id&.to_s,
        key: key,
        level: level,
        channels: channels,
        datas: datas,
        is_slow: is_slow,
      )
    end
  end
end
