# frozen_string_literal: true

require 'rails_helper'

# %w[
#   applicant
#   housing_request
# ].each { |factory_name|require Hermes::Engine.root.join('spec', 'factories', "#{factory_name.pluralize}_factory").to_s }

RSpec.describe Hermes::HousingRequest, type: :model do
  let!(:user) {create(:user)}
  describe 'basic controls' do
    let!(:user1) {create(:user)}
    let!(:applicant1) do
      create(
        :applicant,
        user: user1,
        monthly_charges: {}
      )
    end
    
    it 'has a valid factory' do
      expect(FactoryBot.build(:housing_request, main_applicant: applicant1)).to be_valid
    end
  end
  
  context 'with simple setup' do
    before :each do
      $redis.flushall
    end
    let!(:user1) {create(:user)}
    let!(:user2) {create(:user)}
    let!(:user3) {create(:user)}
    let!(:applicant1) {
      create(
        :applicant,
        user: user1,
        monthly_charges: {},
      )
    }
    let!(:applicant2) {
      create(
        :applicant,
        user: user2
      )
    }
    let!(:applicant3) {
      create(
        :applicant,
        user: user3
      )
    }
    
    let!(:housing_request1) do
      create(
        :complete_request,
        is_active: true,
        main_applicant: applicant1,
        seeked_departments: []
      )
    end
    
    let!(:housing_request2) do
      create(
        :complete_request,
        is_active: true,
        main_applicant: applicant1,
        coapplicants: [
          {rel_type: 'Conjoint (e)', id: applicant2.id.to_s},
          {rel_type: 'Parent (e)', id: applicant3.id.to_s}
        ]
      )
    end
    
    describe 'Eligibility query are valid' do
      it 'should include right typology ranges' do
        typology_range1 = housing_request1.update_typology_range
        typology_range2 = housing_request2.update_typology_range
        range1_to_i = typology_range1.map {|typology| Integer(typology.sub('T', ''))}
        range2_to_i = typology_range2.map {|typology| Integer(typology.sub('T', ''))}
        
        (1..2).each do |n|
          (1..2).each do |i|
            expect(binding.local_variable_get("range#{n}_to_i".to_sym)).to include i
          end
        end
        expect(range1_to_i.max).to be 2
        expect(range2_to_i).to include 3
        expect(range2_to_i.max).to be 3
      end
      
      it 'eligibility should incorpore typology' do
        housing_request1.update_typology_range
        typology_filter1 = housing_request1.typology_filter
        expect(typology_filter1[:typology][:$in]).to include 'T1'
        expect(typology_filter1[:typology][:$in]).to include 'T2'
      end
      
      describe 'edit action dont duplicate coapplicants' do
        before :each do
          housing_request2.coapplicants = [
            {rel_type: 'Conjoint (e)', id: applicant3.id}
          ]
          housing_request2.seeking_elevator = false
          housing_request2.save
          housing_request2.reload
          housing_request2.coapplicants = [{rel_type: 'M', id: applicant3.id.to_s}]
          housing_request2.save
          applicant3.destroy
          housing_request2.reload
          housing_request2.forcing_duplicate = true
          housing_request2.save
          housing_request2.reload
        end
        it 'coapplicants doesnt change' do
          expect(housing_request2.reload.applicant_ids.length).to eq(0)
        end
      end
    end
    
    describe "#update_seeked_location" do
      before :each do
      $redis.flushall
      end
      let(:list) do
        {'nom': 'Fréjus', 'code': '83061', 'codeDepartement': '81', 'codesPostaux': ['83900', '83370', '']}
      end
      let(:lists) do
        [
          {'nom': 'Paris', 'code': '75001', 'codeDepartement': '75', 'codesPostaux': ['75006', nil]},
          {'nom': 'Meaux', 'code': '77000', 'codeDepartement': '77'},
          {'nom': 'Paris', 'code': '75002', 'codeDepartement': '75', 'codesPostaux': ['75006', '93000']}
        ]
      end
      let(:list_without_postal_codes) do
        [
          {'nom': 'Meaux', 'code': '77000', 'codeDepartement': '77'},
        ]
      end

      let(:broken_list) do
        {'codePostal': '77013', 'foo': 'bir'}
      end
      
      it 'sets seeked fields on save' do
        housing_request1.seeked_localisation_list = [list]
        expect {housing_request1.save}.to change {housing_request1.seeked_departments}.from([]).to(['81'])
                                            .and change {housing_request1.truly_seeked_departments}.from(['34']).to(['81', '83'])
                                                   .and change {housing_request1.seeked_district_postal_codes}.from([]).to(['83370', '83900'])
                                                          .and change {housing_request1.seeked_districts}.from([]).to(['Fréjus'])
      end
      
      it 'ensures the sorted uniqueness of seeked fields' do
        housing_request1.seeked_localisation_list = lists + [list]
        housing_request1.save
        expect(housing_request1.seeked_departments).to eq ['75', '77', '81']
        expect(housing_request1.truly_seeked_departments).to eq ['75', '77', '81', '83']
        expect(housing_request1.seeked_district_postal_codes).to eq ['75006', '83370', '83900', '93000']
        expect(housing_request1.seeked_districts).to eq ['Fréjus', 'Meaux', 'Paris']
      end
      
      it 'get truly_seeked_departments from insee code' do
        housing_request1.seeked_localisation_list = list_without_postal_codes
        expect {housing_request1.save}.to change {housing_request1.truly_seeked_departments}.from(["34"]).to(['77'])
      end

      it 'returns error' do
        housing_request1.seeked_localisation_list = [broken_list]
        housing_request1.save
        expect(housing_request1.errors.full_messages.first).to match /non-conform seeked_localisation_list/
      end
    end
    
    describe 'Relationships are valid' do
      before :each do
        $redis.flushall
      end
      
      it 'should has correct values' do
        revenues_sum = (applicant1.monthly_revenues || 0) + (applicant2.monthly_revenues || 0)
        expect(housing_request2.monthly_revenues).to eq(revenues_sum)
      end
      
      it 'should has correct coapplicants' do
        coapplicants = housing_request2.coapplicants
        expect(coapplicants.length).to eq(2)
        expect(coapplicants.map {|coapplicant| coapplicant[:rel_type]}).to include('Conjoint (e)').and include('Parent (e)')
        expect(coapplicants.map {|coapplicant| coapplicant[:id]}).to include(applicant2.id).and include(applicant3.id)
      end
      
      describe 'after coapplicants update' do
        before :each do
          housing_request2.coapplicants = [
            {rel_type: 'Conjoint (e)', id: applicant3.id}
          ]
          housing_request2.save
        end
        
        it 'should has updated with valid data' do
          r_request = housing_request2.reload
          expect(r_request.applicant_ids.length).to be 1
          expect(r_request.applicant_ids.first).to eq applicant3.id
          expect(r_request.coapplicants.first[:rel_type]).to eq 'Conjoint (e)'
        end
      end
    end
    
    describe 'applicant should have only public offer as eligibility when last active request becomes inactive' do
      context 'is_active is set to false' do
        before :each do
          housing_request2.update(is_active: false)
        end
        
        it 'should has is_active == false' do
          expect(housing_request2.reload.is_active).to be false
        end
        
        it 'should has its eligibility == {is_public: true}' do
          expect(housing_request2.reload.eligibility_query[:is_public]).to eq(true)
        end
      end
    end
    
  end

  describe "2 active housing request with same nud is forbidden" do
    let!(:applicant) {create(:applicant)}
    let!(:applicant2) {create(:applicant)}
    let!(:request1) {create(:housing_request, main_applicant: applicant, is_active: true)}
    let!(:request2) {create(:housing_request, main_applicant: applicant2, is_active: false)}
    let!(:nud_test) {create(:nud, housing_request: request1, unique_number: "011752019061716361")}
    let!(:nud_test2) {create(:nud, housing_request: request2, unique_number: "011752019061716361")}
    
    it "should throw error" do
      expect {request2.is_active = true }.to change {request2.valid?}.from(true).to(false)
    end
  end

  describe "generate als id" do 
    let!(:applicant1) {create(:applicant)}
    let!(:applicant2) {create(:applicant)}
    let!(:hr1) { create(:housing_request, main_applicant: applicant1)}
    let!(:hr2) { create(:housing_request, main_applicant: applicant2)}

    it 'job generate 2 als_id' do
      expect( enqueued_jobs.select {|job| job[:job] == AlsIdUniqJob}.length).to eq(2)
      drain_all_jobs
      expect(hr1.reload.als_id).not_to eq(hr2.reload.als_id)
    end 

    it 'cannot save with same als_id' do 
      Hermes::HousingRequest.remove_indexes
      Hermes::HousingRequest.create_indexes
      AlsIdUniqJob.perform_now(hr1.id.to_s, "test")
      AlsIdUniqJob.perform_now(hr2.id.to_s, "test")
      expect(hr1.reload.als_id).to eq("test")
      expect(hr2.reload.als_id).to eq(nil)
    end 

    it 'not regenerate als_id ' do 
      drain_all_jobs
      als_id = '0000000001'
      expect(hr1.reload.als_id).to eq(als_id)
      10.times do |i|
        AlsIdUniqJob.perform_now(hr1.to_s)
      end
      expect(hr1.reload.als_id).to eq(als_id)
    end 

  end 

  describe "check spec via csv" do
    path = [Hermes::Engine.root, %w[spec xml_bouchon], "Nomenclature Cerfa v4 - Sheet1.csv"  ].flatten.join('/').to_s
    only_line= false
    cahier_tests_cerfa_v4 = CSV.read(path)
    cahier_tests_cerfa_v4.each_with_index do |line,i|
      if(line[1] && line[2] && line[3] && line[4] && i>0 && (!only_line || i +1 ==only_line))
        describe "test ligne "+ (i+1).to_s + " : " + line[0] + " ----> check if "+line[1]+" is well extracted in xml with unique number "+line[3] do
          let!(:applicant1) {create(:applicant,
            postal_details: {"district"=> "pantin", "address_insee_code"=> "75056"},
            owned_housing_details: {"district"=> "pantin", "address_insee_code"=> "75056"},
            actual_housing_details: {"district"=> "pantin", "address_insee_code"=> "75056"},

            )}
          let!(:housing_request) { create(:housing_request, main_applicant: applicant1,   
            seeked_localisation_list:[{code: "33318", nom: "valette"}],
            postal_details: {"district"=> "pantin", "address_insee_code"=> "75056"},
            owned_housing_details: {"district"=> "pantin", "address_insee_code"=> "75056"},
            actual_housing_details: {"district"=> "pantin", "address_insee_code"=> "75056"},)}
          let!(:nud) {create(:nud, housing_request: housing_request,  unique_number: line[3])}
          before :each do
            housing_request.update_request_from_sne(user)
          end
          it ('should ave value eq to ' + line[4]) do
            path_in_model = line[1].split(".")
            collected_value=housing_request
            if(housing_request.xml)
              if(path_in_model[0] == "coapplicants")
                seached_applicant = Hermes::Applicant.find(housing_request.coapplicants[path_in_model[1].to_i][:id])
                if(path_in_model[2] == "rel_type")
                  collected_value = housing_request.coapplicants[path_in_model[1].to_i][:rel_type]
                else
                  collected_value = seached_applicant.send(path_in_model[2])
                end
              else
               
                path_in_model.each do |key|
                  if collected_value.respond_to?(key)
                    collected_value = collected_value.send(key)
                  elsif (collected_value[key.to_sym] rescue false)
                    collected_value = collected_value[key.to_sym]
                  elsif (collected_value[key.to_i] rescue false)
                    collected_value = collected_value[key.to_i]
                  else
                    collected_value = collected_value[key.to_s]
                  end
                end
              end
              value_to_compare = line[4]
              case line[4] # a_variable is the variable we want to compare
                when "TRUE"    #compare to 1
                  value_to_compare = true
                when "FALSE"    #compare to 2
                  value_to_compare = false
                when "NIL"    #compare to 2
                  value_to_compare = nil
              end
              if( collected_value.is_a?(Date))
                value_to_compare = Date.parse(line[4])
              end
              if( collected_value.is_a?(Integer))
                value_to_compare = line[4].to_i
              end
              expect(collected_value).to eq(value_to_compare)
            else
              ap "le fichier XML "+ line[3] + " est manquant et ne permet pas de réaliser le test n°"+ i.to_s
            end
          end
        end
      end
    end
  end

  describe "update_from_sne" do
    let!(:applicant1) {create(:applicant)}
    let!(:housing_request) { create(:housing_request, main_applicant: applicant1, user: user, consent_level:1)}
    let!(:nud) {create(:nud, housing_request: housing_request, department: nil,  unique_number: "1110121021518G1104")}

    describe "first test" do 
      before :each do
        housing_request.update_request_from_sne(user)
      end
      it 'should add coapplicants to request' do
        # on check qu'on a le bon nombre d'applicant
        expect(housing_request.coapplicants.length).to eq(8)
        expect(Hermes::Applicant.all.length).to eq(9)
        applicant_id = housing_request.coapplicants[0][:id]
        applicant = Hermes::Applicant.find(applicant_id)
        expect(applicant.nationality).to eq("1")
        #on check qu'ils sont bien updatés
        housing_request.main_applicant.main_employer_name = "minie"
        applicant.nationality = "2"
        applicant.main_employer_name = "picsou"
        applicant.save
        expect(applicant.nationality).to eq("2")

        housing_request.update_request_from_sne(user)
        #on check que les données sont updatées
        expect(applicant.reload.nationality).to eq("1")
        #on check que le main employer name est conservé
        expect(applicant.reload.main_employer_name).to eq("picsou")
        expect(housing_request.main_applicant.reload.main_employer_name).to eq("minie")
        expect(housing_request.consent_level).to eq(1)
        expect(applicant.reload.id.to_s).to eq(applicant_id)
        expect(Hermes::Applicant.all.length).to eq(9)
        applicant.first_name = "momo"
        applicant.last_name = "momo"
        applicant.save

        housing_request.reload.update_request_from_sne(user)
        expect(Hermes::Applicant.where(id: applicant.id).length).to eq(0)
        coloc = Hermes::Applicant.find(housing_request.coapplicants[5][:id])
        housing_request.coapplicants.each do |coapp|
          expect(Hermes::Applicant.find(coapp[:id]).user_id).to eq(user._id)
        end
        # Not working
        # coloc.unemployment_period = true
        # coloc.save
        # housing_request.reload.update_request_from_sne(user)
        # expect(coloc.reload.unemployment_period).to eq(true)
        nud.unique_number = "1110121021545G1104"
        nud.save
        housing_request.reload.update_request_from_sne(user)
        housing_request.coapplicants.each do |coapp|
          applic = Hermes::Applicant.find(coapp[:id])
          expect(applic.user_id).to eq(user._id)
        end
        nud.unique_number = "1110121021518G1104"
        nud.save
        housing_request.reload.update_request_from_sne(user)
        housing_request.coapplicants.each do |coapp|
          applic = Hermes::Applicant.find(coapp[:id])
          expect(applic.user_id).to eq(user._id)
        end

        housing_request.main_applicant.workplace_locality = "pantin"
        housing_request.main_applicant.save
        nud.unique_number = "0331120001997G1175"
        nud.save
        housing_request.reload.update_request_from_sne(user)
          expect(housing_request.reload.main_applicant.workplace_locality).to eq(nil)
       
      end

      it 'update requester_X' do 
        housing_request.update_request_from_sne(user)
        housing_request.reload
        expect(housing_request.requester_firstname).to eq('DONALDO')
        expect(housing_request.requester_lastname).to eq('DUCK')
        expect(housing_request.requester_fullname).to eq('DONALDO DUCK')
        expect(housing_request.requester_email).to eq('d.duck@toto.fr')
        expect(housing_request.requester_birth_date).to eq(Date.parse('1982-10-10'))

      end 
    end
  end

  describe "update_request_supplement" do 
    let!(:user) { create(:user)}
    let!(:user2) { create(:user)}
    let!(:hr_imhoweb) { create(:active_request, user: user)}
    let!(:hr_sne) { create(:active_request, user: user2)}
    let!(:nud_imhoweb) {create(:nud, housing_request: hr_imhoweb, department: nil,  unique_number: "1110121021518G1104")}
    let!(:nud_sne) {create(:nud, housing_request: hr_sne, department: nil,  unique_number: "1110121021532G3U05", provider: 'SNE')}
    let!(:offer) {create(:housing_offer, insee_code: '33000')}

    before :each do
      nud_imhoweb.set(provider: 'IMHOWEB')
      nud_sne.set(provider: 'SNE')
      hr_imhoweb.reload
      hr_sne.reload
    end 

    it 'when SNE' do 
      hr_sne.update_request_from_sne(hr_sne.user)
      expect(hr_sne.reload.cotations).to eq({"94054"=>100})
      expect(hr_sne.reload.supplement_status).to eq(nil)
      expect(hr_sne.reload.supplement_can_apply).to eq(nil)
    end 

    it 'when IMHOWEB' do 
      hr_imhoweb.update_request_from_sne(hr_imhoweb.user)
      expect(hr_imhoweb.reload.cotations).to eq({"44109"=>1000, "44162"=>200})
      expect(hr_imhoweb.reload.supplement_status).to eq("Active")
      expect(hr_imhoweb.reload.supplement_can_apply).to eq(true)
    end 

    it 'when applying SNE' do 
      app = Hermes::HousingApplication.create(housing_request: hr_sne, housing_offer: offer, status: 'EN ATTENTE', visit_status: 'VISITE PROPOSEE')
      expect(hr_sne.reload.supplement_can_apply).to eq(nil)
      expect(app.valid?).to eq(true)
      expect(hr_sne.reload.supplement_status).to eq(nil)
      expect(hr_sne.reload.cotations).to eq(nil)
      expect(hr_sne.reload.supplement_can_apply).to eq(nil)
    end 

    it 'when applying IMHOWEB' do 
      app = Hermes::HousingApplication.create(housing_request: hr_imhoweb, housing_offer: offer, status: 'EN ATTENTE', visit_status: 'VISITE PROPOSEE')
      expect(app.valid?).to eq(true)
      expect(hr_imhoweb.reload.supplement_status).to eq("Active")
      expect(hr_imhoweb.reload.cotations).to eq({"44109"=>1000, "44162"=>200})
      expect(hr_imhoweb.reload.cotations).to eq(app.cotations)
      expect(hr_imhoweb.reload.supplement_can_apply).to eq(true)
    end 

    it 'when applying when cannot' do 
      nud_imhoweb.set(provider: 'IMHOWEB', unique_number: "1110121021518G1105")
      hr_imhoweb.set(unique_numbers: ['1110121021518G1105'])
      app = Hermes::HousingApplication.create(housing_request: hr_imhoweb, housing_offer: offer, status: 'EN ATTENTE', visit_status: 'VISITE PROPOSEE')
      expect(hr_imhoweb.reload.supplement_can_apply).to eq(false)
      expect(hr_imhoweb.reload.cotations).to eq({"44109"=>1000, "44162"=>200})
      expect(app.valid?).to eq(false)
    end 

    it 'when no localisationsSouhaitees' do 
      nud_imhoweb.set(provider: 'IMHOWEB', unique_number: "1110121021518G1106")
      hr_imhoweb.set(unique_numbers: ['1110121021518G1106'])
      hr_imhoweb.update_request_supplement(hr_imhoweb.user)
      expect(hr_imhoweb.reload.cotations).to eq({})
    end 

    it 'when sopra error' do 
      nud_imhoweb.set(provider: 'IMHOWEB', unique_number: "not_found")
      hr_imhoweb.set(unique_numbers: ['not_found'])
      error = hr_imhoweb.update_request_supplement(hr_imhoweb.user)
      expect(hr_imhoweb.reload.cotations).to eq(nil)
      expect(error[:errors]).to eq('missing test file')
    end 

    it 'when applying with timeout' do 
      HERMES_CONFIG['sleeptimeout'] = 11
      app = Hermes::HousingApplication.create(housing_request: hr_imhoweb, housing_offer: offer, status: 'EN ATTENTE', visit_status: 'VISITE PROPOSEE')
      HERMES_CONFIG['sleeptimeout'] = nil

      expect(app.valid?).to eq(true)
      expect(hr_imhoweb.reload.cotations).to eq(nil)
    end 

    it 'update cotations after' do 
      app = Hermes::HousingApplication.create(housing_request: hr_imhoweb, housing_offer: offer, status: 'EN ATTENTE', visit_status: 'VISITE PROPOSEE')
      expect(hr_imhoweb.reload.cotations).to eq({"44109"=>1000, "44162"=>200})
      expect(app.reload.cotations).to eq({"44109"=>1000, "44162"=>200})

      new_cotation = {'test' => 1000}
      hr_imhoweb.update(cotations: new_cotation)
      expect(hr_imhoweb.reload.cotations).to eq(new_cotation)
      expect(app.reload.cotations).to eq(new_cotation)
    end 
    

  end 

  describe "After actions" do
    let!(:user) { create(:user) }
    let!(:lessor) { create(:lessor) }
    let!(:entity) { create(:entity, lessor: lessor, siret: '12345678901234') }
    let!(:housing_offer) { create(:housing_offer, lessor: lessor) }
    let!(:applicant) { create(:applicant, user: user, main_employer_siret: entity.siret) }  
    let!(:housing_request) { create(:housing_request, is_active: true, main_applicant: applicant, request_status: 'COMPLETE') }
    let!(:housing_application) { create(:housing_application, housing_offer: housing_offer, housing_request: housing_request, status: 'TRANSMIS BAILLEUR' ) }
    
    it "should create public shares if request is applicated" do
      hrs = Hermes::HousingRequestShare.find_by(housing_request_id: housing_request.id, lessor_id: lessor.id)
      expect(hrs).not_to be_nil
    end

    it "should create and link housing_request_folder" do
      folders = Hermes::HousingRequestFolder.where(housing_request_id: housing_request.id)
      expect(housing_request.reload.housing_request_folder_id).not_to be_nil
      expect(folders.count).to eq(1)
    end
  end 

  describe "Entity validations" do
    let!(:user) { create(:user) }
    let!(:lessor) { create(:lessor) }
    let!(:entity) { create(:entity, lessor: lessor, siret: '12345678901234') }
    let!(:lessor2) { create(:lessor) }
    let!(:entity2) { create(:entity, lessor: lessor2, auto_request_validation: true, siret: '22345678901234') }
    let!(:applicant) { create(:applicant, user: user, main_employer_siret: entity.siret) }
    let!(:housing_request) { create(:housing_request, is_active: true, main_applicant_id: applicant.id, request_status: 'COMPLETE') }
  
    it "should automaticly validate request" do
      applicant.update(main_employer_siret: entity2.siret)
    
      expect(housing_request.reload.entity_validated).to eq(true)
    end
  
    it "should not automaticly validate request" do
      entity2.update(auto_request_validation: false)
      applicant.update(main_employer_siret: entity2.siret)
    
      expect(housing_request.reload.entity_validated).to eq(false)
    end

  end

  describe "Entity_status workflow" do
    let!(:user) { create(:user) }
    let!(:lessor) { create(:lessor) }
    let!(:entity) { create(:entity, lessor: lessor, siret: '00000000000001') }
    let!(:entity2) { create(:entity, lessor: lessor2, auto_request_validation: false, siret: '00000000000002') }
    let!(:lessor2) { create(:lessor) }
    let!(:applicant) { create(:applicant, user: user, main_employer_siret: entity2.siret, amount_year_minus_two: 200000) }
    let!(:hr) { create(:complete_request, is_active: true, main_applicant: applicant) }
  
    it 'works by default' do 
      expect(hr.reload.entity_status).to eq('A VALIDER')
    end 

    it 'works when entity change autovalidation' do 
      entity2.update(auto_request_validation: true)
      drain_all_jobs
      expect(hr.reload.entity_status).to eq('EN COURS')
    end 

    it 'works when entity_validated' do 
      hr.update(entity_validated: true)
      expect(hr.reload.entity_status).to eq('EN COURS')
    end 
    
  end

  describe 'consent_level 1 and entity points' do 
    let!(:entity) { create(:entity, siret: '00000000000001', spent_on_ongoing_requests: 200)}
    let!(:entity2) { create(:entity, siret: '00000000000002', auto_request_validation: false)}
    let!(:entity_user) { create(:user, group_ids: [entity.manager_group_id])}
    let!(:app) { create(:applicant, birth_date: Date.today - 50.years, main_employer_siret: entity.siret, amount_year_minus_two: 11675,  monthly_incomes: {'Salaire': 1345.45}) }
    let!(:hr) { create(:complete_request, main_applicant: app, is_active: true, seeked_localisation_list: [ {codeDepartement: '34'} ],) }
    let!(:entity_form) { Hermes::HousingRequestEntityForm.create(housing_request_id: hr.id, level: 2)}

    let!(:threshold1) do
      create(
        :threshold,
        type: 'PLA LM',
        thresholds_map: {
          'Paris et communes limitrophes': {
            'thresholds_list': [
              13_050,
              21_272,
              27_883,
              30_521,
              36_307,
              40_859
            ],
            'another_person': 4552
          },
        }
      )
    end
    let!(:offer) { create(:housing_offer,           
      rent_with_charges: 234,
      rent_amount: 234,
      typology: 'T2',
      insee_code: '34000',
      commercialized: false,
      threshold_type: 'PLA LM',
      offer_status: 'LOGEMENT PUBLIE',
      threshold_subtype: 'Paris et communes limitrophes'
    ) }


    before do
      hr.update(
        priority_points_indexes: [10]
      )
    end

    it 'has a link to entity' do
      hr.reload
      app.reload
      entity.reload
      expect(hr.consent_level).to eq(2)
      expect(hr.request_status).to eq('COMPLETE')
      expect(hr.entity_status).to eq('EN COURS')
      expect(hr.entity_id).to eq(entity.id)
      expect(app.readable_by?(entity_user)).to eq true
      expect(hr.readable_by?(entity_user)).to eq true
      expect(Hermes::HousingRequest.where(offer.eligibility_query).pluck(:id).include?(hr.id)).to eq true
      expect(entity.spent_on_ongoing_requests).to eq(210)
      expect(entity.spent_on_completed_requests).to eq(0)
      expect(entity.spent_on_other).to eq(0)
      expect(entity.total_spent).to eq(210)
    end 

    it 'lose link to entity when consent_level == 1' do 
      entity_form.update(level: 1)
      hr.reload
      app.reload
      entity.reload
      expect(hr.consent_level).to eq(1)
      expect(hr.request_status).to eq('COMPLETE')
      expect(hr.entity_status).to eq(nil)
      expect(app.main_employer_siret).to eq(entity.siret)
      expect(hr.old_entity_id).to eq(entity.id)
      expect(hr.entity_id).to eq(nil)
      expect(app.readable_by?(entity_user)).to eq false
      expect(hr.readable_by?(entity_user)).to eq false
      expect(Hermes::HousingRequest.where(offer.eligibility_query).pluck(:id).include?(hr.id)).to eq true
      expect(entity.spent_on_ongoing_requests).to eq(200)
      expect(entity.spent_on_completed_requests).to eq(0)
      expect(entity.spent_on_other).to eq(0)
      expect(entity.total_spent).to eq(200)
    end 

    it 'switch back from consent_level == 1' do 
      entity_form.update(level: 1)
      entity_form.update(level: 3)
      hr.reload
      app.reload
      entity.reload
      expect(hr.consent_level).to eq(3)
      expect(hr.entity_id).to eq(entity.id)
      expect(hr.request_status).to eq('COMPLETE')
      expect(hr.entity_status).to eq('A VALIDER')
      expect(hr.old_entity_id).to eq(entity.id)
      expect(app.readable_by?(entity_user)).to eq true
      expect(hr.readable_by?(entity_user)).to eq true
      expect(Hermes::HousingRequest.where(offer.eligibility_query).pluck(:id).include?(hr.id)).to eq true
      expect(entity.spent_on_ongoing_requests).to eq(200)
      expect(entity.spent_on_completed_requests).to eq(0)
      expect(entity.spent_on_other).to eq(0)
      expect(entity.total_spent).to eq(200)
    end 

    it 'when priority_points_indexes change' do 
      hr.update(priority_points_indexes: [25])
      hr.reload
      app.reload
      entity.reload
      expect(entity.spent_on_ongoing_requests).to eq(225)
      expect(entity.spent_on_completed_requests).to eq(0)
      expect(entity.spent_on_other).to eq(0)
      expect(entity.total_spent).to eq(225)
    end 

    it 'when entity_id_changed change' do 
      app.update(main_employer_siret: entity2.siret)
      hr.reload
      app.reload
      entity.reload
      entity2.reload
      expect(entity.spent_on_ongoing_requests).to eq(200)
      expect(entity.spent_on_completed_requests).to eq(0)
      expect(entity.spent_on_other).to eq(0)
      expect(entity.total_spent).to eq(200)
      expect(entity2.spent_on_ongoing_requests).to eq(0)
      expect(entity2.spent_on_completed_requests).to eq(0)
      expect(entity2.spent_on_other).to eq(0)
      expect(entity2.total_spent).to eq(0)
      expect(hr.priority_points_indexes.sum).to eq(0)
      hr.update(entity_validated: true, priority_points_indexes: [10])
      hr.reload
      app.reload
      entity.reload
      entity2.reload
      expect(entity.spent_on_ongoing_requests).to eq(200)
      expect(entity.spent_on_completed_requests).to eq(0)
      expect(entity.spent_on_other).to eq(0)
      expect(entity.total_spent).to eq(200)
      expect(entity2.spent_on_ongoing_requests).to eq(10)
      expect(entity2.spent_on_completed_requests).to eq(0)
      expect(entity2.spent_on_other).to eq(0)
      expect(entity2.total_spent).to eq(10)
    end 

    (Hermes::HousingRequest.remove_entity_stock_status - ['REFUSE']).each do |type|
      it "entity_status #{type}" do 
        type == 'ANNULE' ? hr.update(request_status: 'ANNULEE') : hr.update(entity_status: type) 
        drain_all_jobs
        hr.reload
        app.reload
        entity.reload
        expect(hr.entity_status).to eq(type)
        expect(entity.spent_on_ongoing_requests).to eq(200)
        expect(entity.spent_on_completed_requests).to eq(0)
        expect(entity.spent_on_other).to eq(0)
        expect(entity.total_spent).to eq(200)
        type == 'ANNULE' ? hr.update(request_status: 'COMPLETE') : hr.update(entity_status: 'EN COURS') 
        drain_all_jobs
        hr.reload
        app.reload
        entity.reload
        expect(entity.spent_on_ongoing_requests).to eq(210)
        expect(entity.spent_on_completed_requests).to eq(0)
        expect(entity.spent_on_other).to eq(0)
        expect(entity.total_spent).to eq(210)
      end 
    end 
  end 

  describe 'Coapplicants Validations' do
    let(:main_applicant) { create(:applicant) }
    let(:housing_request) { create(:housing_request, main_applicant: main_applicant, is_active: true) }
    let(:coapplicant) { create(:applicant) }

    it 'can add coapplicants with blank rel_type' do
      housing_request.update(coapplicants: [{ id: coapplicant.id.to_s, rel_type: '' }])
      expect(housing_request.reload.coapplicants.first[:rel_type]).to be_blank
    end

    it 'can not add coapplicants with blank rel_type if form_status is complete' do
      housing_request.update(form_status: 'complete', coapplicants: [{ id: coapplicant.id.to_s, rel_type: '' }])
      expect(housing_request.errors.any?).to be(true)
      field, error = housing_request.errors.first
      expect(field).to eq(:coapplicants)
      expect(error).to include("rel_type can't be blank for coappllicants")
    end

    it 'can add coapplicants with rel_type if form_status is complete' do
      housing_request.update(form_status: 'complete', coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C' }])
      expect(housing_request.errors.any?).to be(false)
    end
  end

  describe 'typology_range & min_surface' do
    let(:applicant) { create(:applicant, amount_year_minus_two: 50000) }
    let(:coapplicant) { create(:applicant) }
    let(:hr) { create(:housing_request, is_active: true, main_applicant: applicant ) }

    describe 'right for' do 
      pre_mapping = {
        '1 personne seule' => { typology: 1..2, min_surface: 9, less_limited_dependent_nb: 0, coapplicant_nb: 0 },
        '2 personnes' => { typology: 1..3, min_surface: 16, less_limited_dependent_nb: 0, coapplicant_nb: 1 },
        '2 personnes + 1 enfant à charge' => { typology: 2..4, min_surface: 25, less_limited_dependent_nb: 1, coapplicant_nb: 1 },
        '1 personne seule + 1 enfant à charge' => { typology: 2..3, min_surface: 16, less_limited_dependent_nb: 1, coapplicant_nb: 0 },
        '2 personnes + 2 enfants à charge' => { typology: 3..5, min_surface: 34, less_limited_dependent_nb: 2, coapplicant_nb: 1 },
        '1 personne seule + 2 enfants à charge' => { typology: 2..4, min_surface: 25, less_limited_dependent_nb: 2, coapplicant_nb: 0 },
        '2 personnes + 3 enfants à charge' => { typology: 4..6, min_surface: 43, less_limited_dependent_nb: 3, coapplicant_nb: 1 },
        '1 personne seule + 3 enfants à charge' => { typology: 3..5, min_surface: 34, less_limited_dependent_nb: 3, coapplicant_nb: 0 },
        '2 personnes + 4 enfants à charge' => { typology: 4..7, min_surface: 52, less_limited_dependent_nb: 4, coapplicant_nb: 1 },
        '1 personne seule + 4 enfants à charge' => { typology: 4..6, min_surface: 43, less_limited_dependent_nb: 4, coapplicant_nb: 0 },
        '2 personnes + 5 enfants à charge' => { typology: 5..8, min_surface: 61, less_limited_dependent_nb: 5, coapplicant_nb: 1 },
        '1 personne seule + 5 enfants à charge' => { typology: 5..7, min_surface: 52, less_limited_dependent_nb: 5, coapplicant_nb: 0 },
        '2 personnes + 6 enfants à charge' => { typology: 5..9, min_surface: 70, less_limited_dependent_nb: 6, coapplicant_nb: 1 },
        '1 personne seule + 6 enfants à charge' => { typology: 5..8, min_surface: 61, less_limited_dependent_nb: 6, coapplicant_nb: 0 },
        '2 personnes + 7 enfants à charge' => { typology: 5..10, min_surface: 79, less_limited_dependent_nb: 7, coapplicant_nb: 1 },
        '1 personne seule + 7 enfants à charge' => { typology: 5..9, min_surface: 70, less_limited_dependent_nb: 7, coapplicant_nb: 0 },
        '2 personnes + 8 enfants à charge' => { typology: 5..11, min_surface: 88, less_limited_dependent_nb: 8, coapplicant_nb: 1},
        '1 personne seule + 8 enfants à charge' => { typology: 5..9, min_surface: 70, less_limited_dependent_nb: 7, coapplicant_nb: 0 },
        '2 personnes + 9 enfants à charge' => { typology: 5..11, min_surface: 88, less_limited_dependent_nb: 8, coapplicant_nb: 1},
        '1 personne seule + 9 enfants à charge' => { typology: 5..9, min_surface: 70, less_limited_dependent_nb: 7, coapplicant_nb: 0 },
        '2 personnes + 10 enfants à charge' => { typology: 5..11, min_surface: 88, less_limited_dependent_nb: 8, coapplicant_nb: 1},
      }

      mappings = pre_mapping.each do |k, a|
        pre_mapping[k][:typology] = (a[:typology]).to_a.map do |i|
          "T#{i.to_s}" 
        end
        pre_mapping[k][:typology_filter] = 
          (pre_mapping[k][:typology] & ['T6', 'T7', 'T8']).empty? ? pre_mapping[k][:typology] : pre_mapping[k][:typology] + ['T6P']
      end

      mappings.each do |k, mapping|
        context k do 
          it 'get the good values' do
            hr.dependents = Array.new(mapping[:less_limited_dependent_nb], { test: 'test'}) 
            hr.coapplicants = mapping[:coapplicant_nb] == 1 ? [ {id: coapplicant.id, rel_type: 'C'}] : [] 
            expect(hr.update_typology_range).to eq mapping[:typology]
            expect(hr.calculate_min_surface).to eq mapping[:min_surface]
            expect(hr.typology_filter[:typology][:$in]).to eq mapping[:typology_filter]
          end
        end
      end
      
      it 'unborn_children active' do
        birth_date = (DateTime.now + 150.days).to_s
        hr.update!(unborn_children: [
          {"anticipated_birth"=> birth_date, "nb_child_birth"=>"1"},
          {"anticipated_birth"=> birth_date, "nb_child_birth"=>"1"},
        ])
        expect(hr.calculate_unborn_children_nb).to eq 2
        expect(hr.typology_range).to eq mappings['1 personne seule + 2 enfants à charge'][:typology]
        expect(hr.min_surface).to eq mappings['1 personne seule + 2 enfants à charge'][:min_surface]
        expect(hr.eligibility_query[:typology]["$in"]).to eq mappings['1 personne seule + 2 enfants à charge'][:typology]
      end

      it 'unborn_children inactive' do
        birth_date = (DateTime.now + 200.days).to_s
        hr.update(unborn_children: [
          {"anticipated_birth"=> birth_date, "nb_child_birth"=>"1"},
          {"anticipated_birth"=> birth_date, "nb_child_birth"=>"1"},
        ])
        
        expect(hr.calculate_unborn_children_nb).to eq 0
        expect(hr.typology_range).to eq mappings['1 personne seule'][:typology]
        expect(hr.min_surface).to eq mappings['1 personne seule'][:min_surface]
        expect(enqueued_jobs.select {|j| j[:job] == UnbornChildrenJob}.length).to eq 1
        
        Timecop.freeze(DateTime.now + 50.days)
        drain_all_jobs
        expect(hr.calculate_unborn_children_nb).to eq 2
        expect(hr.reload.typology_range).to eq mappings['1 personne seule + 2 enfants à charge'][:typology]
        expect(hr.reload.min_surface).to eq mappings['1 personne seule + 2 enfants à charge'][:min_surface]
        Timecop.return
      end

    end

    describe "inherit birth date from applicant" do
      let!(:applicant) do
        create(
          :applicant,
          birth_date: '2000-01-01'
        )
      end
      let(:housing_request) do
        create(
          :housing_request,
          is_active: true,
          main_applicant: applicant
        )
      end
  
      it 'at creation' do
        expect(housing_request.requester_birth_date).to eq(applicant.birth_date)
      end

      it 'changed applicant birth date' do
        applicant.birth_date = "1900-01-01"
        applicant.save
        applicant.reload
        expect(housing_request.requester_birth_date).to eq(applicant.birth_date)
      end
    end

  end

  describe 'calculate category_type for' do 

    context 'one person' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant)}

      it 'as good category' do 
        expect(hr.reload.category_type).to eq(1)
      end 
    end 

    context 'one persone with unborn_children' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 20.years )}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, unborn_children: [{anticipated_birth: (Date.today + 150.days).to_s, nb_child_birth: 1}] )}
      
      it 'as good category' do 
        expect(hr.reload.young_couple?).to eq(false)
        expect(hr.reload.category_type).to eq(1)
      end 
    end 

    context 'two people (not young couple)' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 30.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 30.years)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [{id: coapplicant.id.to_s, rel_type: 'C'}])}
      
      it 'as good category' do 
        expect(hr.reload.category_type).to eq(2)
      end 
    end 
    
    context 'one person with handicap' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, disability_details: [{lol: 'test'}])}
      
      it 'as good category' do 
        expect(hr.reload.handicap?).to eq(true)
        expect(hr.reload.category_type).to eq(2)
      end 
    end 

    context 'single_parent with visit' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, dependents: [{rel_type: 'D'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(false)
        expect(hr.reload.category_type).to eq(2)
      end 
    end 

    context '3 people' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 30.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 30.years )}
      let(:coapplicant2) { create(:applicant, birth_date: DateTime.now - 30.years )}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [{id: coapplicant.id.to_s, rel_type: 'C'}, {id: coapplicant2.id.to_s, rel_type: 'C'}])}
      
      it 'as good category' do 
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context '1 person with on dependent' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, dependents: [{rel_type: 'E'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context '1 person with on dependent parent' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, dependents: [{coparentality: 'P'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(false)
        expect(hr.reload.category_type).to eq(2)
      end 
    end 

    context 'young couple' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 20.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 20.years)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C'}])}
      
      it 'as good category' do 
        expect(hr.reload.young_couple?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context '2 people with handicap' do 
      let(:applicant) { create(:applicant)}
      let(:coapplicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [ {id: coapplicant.id.to_s, rel_type: 'C'} ], disability_details: [{lol: 'test'}])}
      
      it 'as good category' do 
        expect(hr.reload.handicap?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context '4 people' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 30.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 30.years )}
      let(:coapplicant2) { create(:applicant, birth_date: DateTime.now - 30.years )}
      let(:coapplicant3) { create(:applicant, birth_date: DateTime.now - 30.years )}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [{id: coapplicant.id.to_s, rel_type: 'C'}, {id: coapplicant2.id.to_s, rel_type: 'C'}, {id: coapplicant3.id.to_s, rel_type: 'C'}])}
      
      it 'as good category' do 
        expect(hr.reload.category_type).to eq(4)
      end 
    end 

    context '1 person with 2 dependents' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, dependents: [{rel_type: 'E'}, {rel_type: 'enfant à charge'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(true)
        expect(hr.reload.category_type).to eq(4)
      end 
    end 

    context '1 person with 2 dependents + one visit' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, dependents: [{rel_type: 'E'}, {rel_type: 'enfant à charge'}, {rel_type: 'D'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(true)
        expect(hr.reload.category_type).to eq(5)
      end 
    end 

    context '2 person with 2 dependents' do 
      let(:applicant) { create(:applicant)}
      let(:coapplicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, coapplicants: [{id: coapplicant.id.to_s, rel_type: 'C'}], main_applicant: applicant, dependents: [{rel_type: 'G'}, {rel_type: 'enfant à charge'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(false)
        expect(hr.reload.category_type).to eq(4)
      end 
    end 

    context '2 person with 2 dependents + one visit' do 
      let(:applicant) { create(:applicant)}
      let(:coapplicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, coapplicants: [ {id: coapplicant.id.to_s, rel_type: 'C'}], main_applicant: applicant, dependents: [{rel_type: 'D'}, {rel_type: 'enfant à charge'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(false)
        expect(hr.reload.category_type).to eq(4)
      end 
    end 

    context '3 people with handicap' do 
      let(:applicant) { create(:applicant)}
      let(:coapplicant) { create(:applicant)}
      let(:coapplicant2) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [ {id: coapplicant.id.to_s, rel_type: 'C'}, {id: coapplicant2.id.to_s, rel_type: 'C'} ], disability_details: [{lol: 'test'}])}
      
      it 'as good category' do 
        expect(hr.reload.handicap?).to eq(true)
        expect(hr.reload.category_type).to eq(4)
      end 
    end 

    context '1 person with 3 dependents' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, dependents: [{rel_type: 'E'}, {rel_type: 'enfant à charge'}, {rel_type: 'garde alternée'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(true)
        expect(hr.reload.category_type).to eq(5)
      end 
    end 

    ##########################################

    context 'young couple with 1 children' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 20.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 20.years)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C'}], dependents: [{rel_type: 'E'}] )}
      
      it 'as good category' do 
        expect(hr.reload.young_couple?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context 'young couple with 1 parent' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 20.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 20.years)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C'}], dependents: [{rel_type: 'P'}] )}
      
      it 'as good category' do 
        expect(hr.reload.young_couple?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context 'young couple with handicap' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 20.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 20.years)}
      let!(:hr) { create(:housing_request,is_active: true, main_applicant: applicant, coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C'}], disability_details: [{lol: 'test'}] )}
      
      it 'as good category' do 
        expect(hr.reload.young_couple?).to eq(true)
        expect(hr.reload.handicap?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context 'young couple with changing date' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 20.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 40.years)}
      let!(:hr) { create(:housing_request, is_active: true,  main_applicant: applicant, coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C'}] )}
      
      it 'as good category' do 
        expect(hr.reload.young_couple?).to eq(false)
        expect(hr.reload.category_type).to eq(2)
        coapplicant.update(birth_date: DateTime.now - 20.years)
        expect(hr.reload.young_couple?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 
  end 

  describe 'inheriht reference_tax_revenues from coaplicants' do
    let!(:applicant) {create(:applicant)}
    let!(:coapplicant) {create(:applicant)}
    let!(:coapplicant2) {create(:applicant)}
    let!(:hr) {create(:housing_request, main_applicant: applicant, coapplicants: [{id: coapplicant.id.to_s, rel_type: 'C'}, {id: coapplicant2.id.to_s, rel_type: 'C'}], is_active: true, form_status: 'invalid', seeked_departments: ['34'])}
    let!(:nud) { create(:nud,housing_request: hr, unique_number: '1')}

    it 'without amount_year_minus_two' do
      expect(hr.reload.reference_tax_revenues).to eq(nil)
    end

    it 'without amount_year_minus_two 0' do
      applicant.update(amount_year_minus_two: 0)
      expect(hr.reload.reference_tax_revenues).to eq(0)
    end

    it 'only applicant amount_year_minus_two' do
      applicant.update(amount_year_minus_two: 1000)
      expect(hr.reload.reference_tax_revenues).to eq(1000)
    end
    
    it 'all applicants amount_year_minus_two' do
      applicant.update(amount_year_minus_two: 1000)
      coapplicant.update(amount_year_minus_two: 2000)
      expect(hr.reload.reference_tax_revenues).to eq(3000)
      coapplicant2.update(amount_year_minus_two: 500)
      expect(hr.reload.reference_tax_revenues).to eq(3500)
    end

    it 'coapplicant affected after hr creation' do
      coapp = Hermes::Applicant.create(amount_year_minus_two: 800, year_minus_two: Date.today.year - 2)
      hr.update(coapplicants: [{id: coapp.id, rel_type: 'C'}])
      expect(hr.reload.reference_tax_revenues).to eq(800)
    end

    it 'request_status changed' do
      hr.validate_all_required_document
      old_truly_seeked_departments= hr.truly_seeked_departments
      hr.update(truly_seeked_departments: [])
      expect(hr.reload.reference_tax_revenues).to eq(nil)
      expect(hr.reload.request_status).to eq("INCOMPLETE")
      
    end
  end

  describe 'housing_channel_keys' do 
    let!(:dalo){ create(:housing_channel, key: 'dalo')}
    let!(:lol){ create(:housing_channel, key: 'lol')}
    let!(:applicant){ create(:applicant)}
    let!(:request1) { create(:housing_request, main_applicant: applicant)}
    let!(:request2) { create(:housing_request, main_applicant: applicant, housing_channel_keys: ['dalo'])}

    it 'create with and without housing_channel_keys' do 
      expect(request1.housing_channel_keys).to eq(['dc'])
      expect(request2.housing_channel_keys).to eq(['dalo', 'dc'])
    end 

    it 'when delete housing_channel_keys' do 
      request2.update(housing_channel_keys: [])
      expect(request2.reload.housing_channel_keys).to eq(['dc'])
    end 

    it 'when add housing_channel_keys' do 
      request1.update(housing_channel_keys: ['dalo'])
      expect(request1.reload.housing_channel_keys).to eq(['dalo', 'dc'])
    end 
  end 

  describe 'update_readers_writers' do
    let!(:entity) {create(:entity, siret: '00000000000001')} 
    let!(:entity_user){ create(:user)}
    let!(:user_right) { create(:user_right, type: 'entity', entity: entity, user: entity_user, right: 'administrator')}
    let!(:applicant){ create(:applicant, user: user)}
    let!(:coapp){ create(:applicant, user: user)}
    let!(:request) { create(:housing_request, main_applicant: applicant, is_active: true, user: user, coapplicants: [{id: coapp.id, rel_type: 'C'}])}

    it 'setup default' do
      readers = []
      writers = [request.user_id.to_s]
      [applicant.reload, coapp.reload, request.reload].each do |r|
        expect(r.readers).to eq(readers)
        expect(r.writers).to eq(writers)
        expect(request.readable_by?(entity_user)).to eq(false)
      end
    end 

    it 'when entity change' do
      readers = []
      writers = [request.user_id.to_s]
      readers_with_entity = [ "g_" + entity.group_id]
      writers_with_entity = writers + [entity.supervisor_group_id, entity.manager_group_id].map {|id| "g_" + id.to_s}

      applicant.update(main_employer_siret: entity.siret)
      [applicant.reload, coapp.reload, request.reload].each do |r|
        expect(r.readers).to eq(readers_with_entity)
        expect(r.writers).to eq(writers_with_entity)
        expect(request.readable_by?(entity_user)).to eq(true)
      end

      applicant.update(main_employer_siret: nil)
      [applicant.reload, coapp.reload, request.reload].each do |r|
        expect(r.readers).to eq(readers)
        expect(r.writers).to eq(writers)
        expect(request.readable_by?(entity_user)).to eq(false)
      end
    end 
  end 

  describe 'update_request_from_sne update dalo and housing_channel_keys' do
    let!(:dalo){ create(:housing_channel, key: 'dalo')}
    let!(:applicant) {create(:applicant)}
    let!(:hr_with_dalo) {create(:housing_request, main_applicant: applicant)}
    let!(:nud1) { create(:nud,housing_request: hr_with_dalo, unique_number: '1110121047572GDPUB')}
    let!(:hr_without_dalo) {create(:housing_request, main_applicant: applicant)}
    let!(:nud2) { create(:nud,housing_request: hr_without_dalo, unique_number: '1110121021545G1104')}

    it 'dont import dalo' do 
      hr_with_dalo.update_request_from_sne(user)
      hr_with_dalo.reload
      expect(hr_with_dalo.dalo).to eq(false)
      expect(hr_with_dalo.housing_channel_keys).to eq([ 'dc'])
    end 

    it 'not remove dalo' do 
      hr_with_dalo.update(housing_channel_keys: ['dc', 'dalo'])
      hr_with_dalo.update_request_from_sne(user)
      hr_with_dalo.reload
      expect(hr_with_dalo.dalo).to eq(true)
      expect(hr_with_dalo.housing_channel_keys).to eq(['dc', 'dalo'])
    end 
    
  end 

  describe 'update_request_from_sne dynamic' do 
    let!(:applicant) {create(:applicant)}
    let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant)}
    let!(:nud) {create(:nud, unique_number: '1110121021545G1104', housing_request: hr)}
    let!(:setting) { create(:setting, key: 'hermes_rfr_settings', values: { change_minus_two_date: Date.today.beginning_of_year.to_s, change_minus_two_year: Date.today.year - 3}) } 

    before do
      Timecop.freeze(Time.local(2021, 2, 1, 0, 0))
      hr.update_request_from_sne(nil)
      hr.validate_all_required_document
      hr.set(request_status: 'COMPLETE')
      Timecop.return
    end 

    it 'default' do 
      hr.reload
      hr.housing_request_folder.reload
      expect(hr.housing_request_folder.consultation_status).to eq('Dossier validé')
      expect(hr.housing_request_folder.application_status).to eq('Dossier validé')
      expect(hr.reference_tax_revenues).to eq(22000)
      expect(hr.request_status).to eq('COMPLETE')
    end 

    it 'with config same year' do 
      date = Date.new(2021, 4, 1,)
      setting.update(values: { change_minus_two_date:  date + 10.days, change_minus_two_year: date.year - 1 })  
      Timecop.freeze(date)
      hr.update_request_from_sne(nil)
      hr.reload
      hr.housing_request_folder.reload
      expect(hr.reference_tax_revenues).to eq(22000)
      expect(hr.housing_request_folder.consultation_status).to eq('Dossier incomplet')
      expect(hr.housing_request_folder.application_status).to eq('Dossier incomplet')
      expect(hr.housing_request_folder.get_not_declared.all? { |type| type.include?('Bulletin')}).to eq(true)
      Timecop.return
    end 
    
    it 'with config next year' do 
      date = Date.new(2021, 4, 1,)
      setting.update(values: { change_minus_two_date: date - 10.days, change_minus_two_year: date.year - 1 })  
      Timecop.freeze(date)
      hr.update_request_from_sne(nil)
      hr.reload
      hr.housing_request_folder.reload
      expect(hr.reference_tax_revenues).to eq(22500)
      expect(hr.housing_request_folder.consultation_status).to eq('Dossier incomplet')
      expect(hr.housing_request_folder.application_status).to eq('Dossier incomplet')
      Timecop.return
    end

    it 'when no value' do 
      date = Date.new(2050, 4, 1,)
      setting.update(values: { change_minus_two_date: date - 10.days, change_minus_two_year: date.year - 1 })  
      Timecop.freeze(date)
      hr.update_request_from_sne(nil)
      hr.reload
      hr.housing_request_folder.reload
      expect(hr.reference_tax_revenues).to eq(nil)
      expect(hr.form_sne_validation_errors.any? { |e| e['hermes_field'] == "main_applicant.dynamic_amount_minus_two" }).to eq(true)
      Timecop.return
    end 

  end 

  describe 'update_request_from_sne with multiple dependents monthly incomes' do
    let!(:applicant) {create(:applicant)}
    let!(:hr) {create(:housing_request, main_applicant: applicant)}
    let!(:nud1) { create(:nud,housing_request: hr, unique_number: '1110121047572GDPUB')}

    it 'works' do 
      hr.update_request_from_sne(user)
      hr.reload
      expect(hr.dependents.first['monthly_incomes'].length).to eq(15)
      expect(hr.dependents.first['monthly_charges'].length).to eq(2)
      expect(hr.dependents_monthly_revenues).to eq(790)
    end 
  end 

  describe 'update_request_from_sne other_disability' do 
    let!(:hr_without_other) {create(:complete_request)}
    let!(:nud1) { create(:nud,housing_request: hr_without_other, unique_number: '1110121047572GDPUB')}
    let!(:hr_with_other) {create(:complete_request)}
    let!(:nud2) { create(:nud,housing_request: hr_with_other, unique_number: '1110121047572GDPUC')}

    it 'works' do 
      hr_without_other.update_request_from_sne(user)
      expect(hr_without_other.reload.disability_details.first[:other_disability]).to eq(false)
      hr_with_other.update_request_from_sne(user)
      expect(hr_with_other.reload.disability_details.first[:other_disability]).to eq(true)
    end 

  end 

  describe 'score' do 
    let!(:entity) {create(:entity, siret: '00000000000001', auto_request_validation: false)}
    let!(:hr) {create(:complete_request)}
    let!(:applicant) { hr.main_applicant }
    let!(:offer) {create(:housing_offer)}
    let!(:ha) { create(:housing_application, status: 'EN ATTENTE', housing_request: hr, housing_offer: offer)}

    it 'work with minimium info' do 
      hr.reload
      expect(hr.profile_score).to eq 0
      expect(ha.profile_score).to eq hr.profile_score
    end

    it 'when consent_level 1' do 
      applicant.update(main_employer_siret: entity.siret)
      hr.reload.update(priority_points_indexes: [100], entity_validated: true, consent_level: 2)
      expect do
        hr.update(consent_level: 1)
      end.to change { hr.reload.profile_score }.from(340).to(0)
      expect(hr.priority_points_indexes).to eq([])
      expect(hr.entity_validated).to eq(false)
    end 
    
    it 'update ha score when score change' do 
      expect(ha.reload.profile_score).to eq 0
      hr.update(sne_created_at: DateTime.now - 30.months)
      expect(hr.reload.profile_score).to eq 54
      expect(ha.reload.profile_score).to eq 54
    end 

    it 'seniority_points check' do 
      hr.update(sne_created_at: DateTime.now - 11.months)
      hr.reload
      expect(hr.profile_score).to eq 11
      expect(hr.score_details['seniority_points']).to eq 11

      hr.update(sne_created_at: DateTime.now - 18.months)
      hr.reload
      expect(hr.profile_score).to eq 24
      expect(hr.score_details['seniority_points']).to eq 24

      hr.update(sne_created_at: DateTime.now - 32.months)
      hr.reload
      expect(hr.profile_score).to eq 60
      expect(hr.score_details['seniority_points']).to eq 60
    end 
    
    it 'under_30_year_points check' do 
      birth_date = Date.today - 20.years
      applicant.update(birth_date: birth_date)
      hr.reload
      expect(hr.requester_birth_date).to eq birth_date
      expect(hr.profile_score).to eq 70
      expect(hr.score_details['under_30_year_points']).to eq 70
    end 

    it 'bonus_point check' do 
      hr.update(bonus_points: 13)
      expect(hr.reload.profile_score).to eq 13
      expect(hr.score_details['bonus_points']).to eq 13
    end 

    it 'priority_points check' do 
      applicant.update(main_employer_siret: entity.siret)
      hr.reload.update(priority_points_indexes: [100], entity_validated: true, consent_level: 2)
      expect(hr.profile_score).to eq 340
      expect(hr.score_details['priority_points']).to eq 100

      applicant.update(main_employer_siret: nil)
      hr.reload
      expect(hr.profile_score).to eq 0
      expect(hr.score_details['priority_points']).to eq 0
    end 

    it 'motivations check' do 
      hr.update(motivations: ['19', '28', '06'])
      hr.reload
      expect(hr.profile_score).to eq 120+70+40
      expect(hr.score_details['professional_mobility_points']).to eq 70
      expect(hr.score_details['closer_to_workplace_points']).to eq 120
      expect(hr.score_details['violence_points']).to eq 40
    end 

    it 'workplace_insee_code_points check' do 
      offer.set(insee_code: '33000')
      hr.main_applicant.set(workplace_insee_code: '33000')
      hr.calculate_score
      hr.save

      hr.reload
      expect(hr.profile_score).to eq 40
      expect(hr.score_details['workplace_insee_code_points']).to eq 40

      ha.update(status: 'ANNULEE', is_active: false)

      hr.reload
      expect(hr.profile_score).to eq 0
      expect(hr.score_details['workplace_insee_code_points']).to eq 0
    end 

    it 'handicap_points check' do 
      Mongoid::Tasks::Database.remove_undefined_indexes
      Hermes::HousingRequest.create_indexes
      Hermes::HousingOffer.create_indexes

      hr.update(disability_details: [{ sensory_disability_kinds: ['DA']}])
      hr.reload
      expect(hr.profile_score).to eq 40
      expect(hr.score_details['handicap_points']).to eq 40

      hr.update(disability_details: [{ physical_disability_kinds: ['TEST'], other_disability: false}])
      hr.reload
      expect(hr.profile_score).to eq 0
      expect(hr.score_details['handicap_points']).to eq 0

      hr.update(disability_details: [{ other_disability: true}])
      hr.reload
      expect(hr.profile_score).to eq 40
      expect(hr.score_details['handicap_points']).to eq 40
    end 

    it 'application_rejected_points check' do 
      ha.update(status: 'REFUSEE EN CAL')
      hr.reload
      expect(hr.profile_score).to eq 40
      expect(hr.score_details['application_rejected_points']).to eq 40
      
      hr.update(application_rejected_nb: 21)
      hr.reload
      expect(hr.profile_score).to eq 840
      expect(hr.score_details['application_rejected_points']).to eq 840
    end 

    it 'application_not_selected_points check' do 
      ha.update(status: 'NON RETENUE')
      expect(hr.profile_score).to eq 5
      expect(hr.score_details['application_not_selected_points']).to eq 5

      hr.update(application_not_selected_nb: 21)
      
      expect(hr.profile_score).to eq 100
      expect(hr.score_details['application_not_selected_points']).to eq 100
    end 

    it 'application_not_selected_points not work when wrong status' do 
      ha.update(status: 'INSCRIPTION CAL')
      ha.update(status: 'NON RETENUE')
      hr.reload
      expect(hr.profile_score).to eq 0
      expect(hr.score_details['application_not_selected_points']).to eq 0
    end 

    it 'application_taken_back_points check' do 
      ha.update(status: 'REPRIS BAILLEUR')
      expect(hr.profile_score).to eq 5
      expect(hr.score_details['application_taken_back_points']).to eq 5

      hr.update(application_taken_back_nb: 21)
      
      expect(hr.profile_score).to eq 100
      expect(hr.score_details['application_taken_back_points']).to eq 100
    end 

    it 'when not in progress and not already_had_score' do 
      hr.update(nud_ids: [], unique_numbers: [], already_had_score: false)
      hr.update(motivations: ['19'])
      expect(hr.reload.request_status).to eq "EN ATTENTE DE NUD"

      expect(hr.reload.profile_score).to eq 0
      expect(ha.reload.profile_score).to eq 0
    end 

    it 'when not in progress and already_had_score' do 
      hr.update(nud_ids: [], unique_numbers: [])
      hr.update(motivations: ['19'])

      expect(hr.reload.profile_score).to eq 120
      expect(ha.reload.profile_score).to eq 120
    end 

    it 'when attente de piece' do 
      hr.housing_request_folder.required_documents.first.update(status: "Pièce manquante", declared: false)
      hr.update(sne_created_at: DateTime.now - 30.months)
      expect(hr.reload.request_status).to eq('EN ATTENTE DE PIECES')
      expect(hr.reload.profile_score).to eq 54
    end   

    describe 'entity_validated_points' do 

      before do 
        applicant.update(main_employer_siret: entity.siret)
        hr.reload.update(consent_level: 2)
      end 

      it 'manual validation' do 
        expect do
          hr.reload.update(entity_validated: true)
        end.to change { hr.reload.score_details['entity_validated_points']}.from(0).to(240)
      end 

      it 'auto_request_validation' do 
        applicant.update(main_employer_siret: nil)
        entity.update(auto_request_validation: true)
        expect do
          applicant.update(main_employer_siret: entity.siret)
          hr.reload.update(consent_level: 2)
        end.to change { hr.reload.score_details['entity_validated_points']}.from(0).to(240)
      end 
  
      it 'auto_request_validation changed' do 
        entity.update(auto_request_validation: false)

        expect do
          entity.update(auto_request_validation: true)
          drain_all_jobs
        end.to change { hr.reload.score_details['entity_validated_points']}.from(0).to(240)
      end 

      describe 'with auto_request_validation' do 
        before do 
          entity.update(auto_request_validation: true)
          drain_all_jobs
        end 

        [1, nil].each do |type|
          it "consent_level #{type}" do 
            hr.reload.update(priority_points_indexes: [10])
            expect do
              hr.reload.update(consent_level: 1)
            end.to change { hr.reload.profile_score }.from(250).to(0)
          end 
        end 

        it 'entity_validation removed' do 
          hr.reload.update(priority_points_indexes: [10])

          expect do
            hr.reload.update(entity_validated: false)
          end.to change { hr.reload.profile_score }.from(250).to(0)
        end 

      end 

      describe 'as applicant_used_for_siret' do 
        let!(:app2) {create(:applicant, main_employer_siret: entity.siret, work_contract_type: 'CDIP')}

        before do 
          entity.update(auto_request_validation: true)
          applicant.update(main_employer_siret: nil)
        end 

        it 'when applicant_used_for_siret change' do
          expect do
            hr.reload.update(applicant_used_for_siret: app2.id, consent_level: 2)
          end.to change { hr.reload.score_details['entity_validated_points']}.from(0).to(240)
        end 
      end 
    end 
  end 

  describe 'calculate_next_seniority_updated_date' do 
    let!(:applicant) {create(:applicant)}
    let!(:hr) {create(:housing_request, main_applicant: applicant, is_active: true)}

    [
      {
        title: 'for same day',
        current_date: DateTime.new(2020, 1, 2),
        sne_created_at_date: DateTime.new(2019, 7, 2),
        expected_date: DateTime.new(2020, 2, 3),
        expected_seniority_point: 6,
      },      
      {
        title: 'for same month date',
        current_date: DateTime.new(2020, 1, 2),
        sne_created_at_date: DateTime.new(2019, 7, 15),
        expected_date: DateTime.new(2020, 1, 16),
        expected_seniority_point: 5,
      },
      {
        title: 'for next month',
        current_date: DateTime.new(2020, 1, 22),
        sne_created_at_date: DateTime.new(2019, 7, 15),
        expected_date: DateTime.new(2020, 2, 16),
        expected_seniority_point: 6,
      },
      {
        title: 'for not existing date',
        current_date: DateTime.new(2020, 2, 2),
        sne_created_at_date: DateTime.new(2019, 7, 30),
        expected_date: DateTime.new(2020, 3, 1),
        expected_seniority_point: 6,
      },
    ].each do |d|
      it d[:title] do 
        Timecop.freeze(d[:current_date])
        hr.set(sne_created_at: d[:sne_created_at_date])
        expect(hr.calculate_next_seniority_updated_date).to eq(d[:expected_date].utc)
        expect(hr.calculate_seniority_points).to eq(d[:expected_seniority_point])
        Timecop.return
        Timecop.freeze(d[:expected_date])
        expect(hr.calculate_seniority_points).to eq(d[:expected_seniority_point] + 1)
        Timecop.return
      end 
    end 
  end 

  describe 'update_renewal_values' do 
    let!(:created_at) { Date.new(2020, 1, 1)}
    let!(:applicant) {create(:applicant,)}
    let!(:hr) {create(:housing_request, main_applicant: applicant, created_at: created_at)}

    it 'works without sne_created_at' do 
      expect(hr.reload.next_sne_renewal_date).to eq(created_at + 1.year)
    end 

    it 'works with sne_created_at' do 
      sne_created_at = Date.new(2020, 4, 1)
      expect do
        hr.update(sne_created_at: sne_created_at)
      end.to change { hr.reload.next_sne_renewal_date }.from(created_at + 1.year).to(sne_created_at + 1.year)
    end 

    it 'works when last_sne_renewal_date changed' do 
      last_sne_renewal_date = Date.new(2020, 6, 1)
      expect do
        hr.update(last_sne_renewal_date: last_sne_renewal_date)
      end.to change { hr.reload.next_sne_renewal_date }.from(created_at + 1.year).to(last_sne_renewal_date + 1.year)
    end 

    it 'update renewal bolean' do 
      last_sne_renewal_date = Date.today - 350.days
      
      hr.update(last_sne_renewal_date: Date.today - 6.months)
      expect(hr.reload.soon_need_to_be_renewed).to eq(false)
      expect(hr.reload.need_to_be_renewed).to eq(false)

      hr.update(last_sne_renewal_date: last_sne_renewal_date)
      expect(hr.reload.soon_need_to_be_renewed).to eq(true)
      expect(hr.reload.need_to_be_renewed).to eq(false)
 
      hr.update(last_sne_renewal_date: last_sne_renewal_date - 1.month)
      expect(hr.reload.soon_need_to_be_renewed).to eq(false)
      expect(hr.reload.need_to_be_renewed).to eq(true)
    end

    it 'force when next_renewal_date_set_by_sne' do 
      expect do
        hr.update(next_renewal_date_set_by_sne: created_at + 5.days)
      end.to change { hr.reload.next_sne_renewal_date }.from(created_at + 1.year).to(created_at + 5.days)
    end 

  end 

  describe 'update_renewed_booleans' do 
    let!(:today) { Date.today}
    let!(:applicant1) {create(:applicant)}
    let!(:hr1) {create(:housing_request, main_applicant: applicant1, created_at: today)}
    let!(:applicant2) {create(:applicant)}
    let!(:hr2) {create(:housing_request, main_applicant: applicant2, created_at: today - 20.days)}
    let!(:applicant3) {create(:applicant)}
    let!(:hr3) {create(:housing_request, main_applicant: applicant3, created_at: today - 35.days)}


    it 'works' do 
      expect(hr1.reload.soon_need_to_be_renewed).to eq(false)
      expect(hr1.reload.need_to_be_renewed).to eq(false)
      expect(hr2.reload.soon_need_to_be_renewed).to eq(false)
      expect(hr2.reload.need_to_be_renewed).to eq(false)
      expect(hr3.reload.soon_need_to_be_renewed).to eq(false)
      expect(hr3.reload.need_to_be_renewed).to eq(false)

      Timecop.freeze(today + 11.month - 1.days)
      Hermes::HousingRequest.update_renewed_booleans
      
      expect(hr1.reload.soon_need_to_be_renewed).to eq(false)
      expect(hr1.reload.need_to_be_renewed).to eq(false)
      expect(hr2.reload.soon_need_to_be_renewed).to eq(true)
      expect(hr2.reload.need_to_be_renewed).to eq(false)
      expect(hr3.reload.soon_need_to_be_renewed).to eq(false)
      expect(hr3.reload.need_to_be_renewed).to eq(true)

      Timecop.freeze(today + 12.month)
      Hermes::HousingRequest.update_renewed_booleans
      expect(hr2.reload.soon_need_to_be_renewed).to eq(false)
      expect(hr2.reload.need_to_be_renewed).to eq(true)
      Timecop.return

    end 
  end 

  describe 'check common validations' do 
    let(:hr) {create(:complete_request)}
    let(:applicant) { hr.main_applicant }
    let(:entity) {create(:entity, siret: '12345678901234')}
   
    subject { hr.reload.extract_errors_lines }

    it 'no errors' do 
      expect(subject.empty?).to be_truthy
    end 

    it 'main_applicant errors' do
      applicant.update(
        first_name: nil,
        last_name: nil,
        birth_date: nil,
        nationality: nil,
        civility: nil,
        email: nil,
        postal_details: {etranger: true},
        actual_housing_details: {},
        work_contract_type: nil,
        amount_year_minus_two: nil,
      )

      expect(subject).to match_array([
        "Situation personnelle Demandeur Civilité",
        "Situation personnelle Demandeur Prénoms",
        "Situation personnelle Demandeur Date de naissance",
        "Situation personnelle Demandeur Nationalité",
        "Situation personnelle Demandeur Mail personnel",
        "Logement actuel Etes-vous  propriétaire d'un logement autre que celui que vous habitez ?  (ou votre conjoint ou le futur co-titulaire du bail)",
        "Situation professionnelle Demandeur Statut",
        "Situation personnelle Adresse courrier Voie",
        "Situation personnelle Adresse courrier Commune",
        "Ressources actuelles Demandeur Avis d'imposition 2022 sur les revenus de l'année 2021",
      ])
    end

    it 'housing_request errors' do
      hr.update(
        seeked_types: [],
        seeked_localisation_list: [],
        widen_request: nil,
      )

      expect(subject).to match_array([
        "Motifs et logement recherché Type de logement",
        "Motifs et logement recherché Localisations souhaitées",
        "Motifs et logement recherché Acceptez-vous que votre demande soit élargie aux autres communes de l'agglomération ? (communauté urbaine ou d'agglomération ou de communes)",
      ])
    end 

    it 'dependents' do 
      hr.update(dependents: [{ rel_type: ['LOLOL'] }])

      expect(subject).to match_array([
        "Situation personnelle Personnes à charge Lien avec le demandeur",
        "Situation personnelle Gardes/Visites Avez-vous un enfant en garde alternée > Date de naissance",
        "Situation personnelle Gardes/Visites Avez-vous un enfant en droit de visite > Date de naissance",
      ])
    end 

    it 'unborn_children' do 
      hr.update(unborn_children: [{ anticipated_birth: (DateTime.now + 150.days).to_s,  nb_child_birth: 1}, { test: 'lol'}])

      expect(subject).to match_array([
        "Situation personnelle Enfant à naître Vous attendez un enfant > Date de naissance prévue"
      ])
    end 

    it 'coapplicants' do 
      app2 = create(:applicant, work_contract_type: nil)
      app3 = create(:applicant, work_contract_type: 'CDI', main_employer_siret: nil, workplace_insee_code: nil)

      hr.update(
        coapplicants: [{id: app2.id.to_s, rel_type: 'C'}, { id: app3.id.to_s, rel_type: 'C'}],
      )

      expect(subject).to match_array([
        "Situation professionnelle Co-demandeur Statut",
        "Situation professionnelle Co-demandeur Ville du lieu de travail",
        "Situation professionnelle Co-demandeur Siret",
        "Ressources actuelles Co-Demandeur Avis d'imposition 2022 sur les revenus de l'année 2021",
      ])
    end 

    context 'main_employer_name' do 

      it 'main_employer_siret not blank with name' do 
        hr.get_applicant_used_for_siret.update(main_employer_name: 'test', main_employer_siret: '123')
        hr.reload.update(consent_level: 3)

        expect(subject).to match_array([])
      end 
      
      it 'main_employer_siret not blank without name' do 
        hr.get_applicant_used_for_siret.update(main_employer_name: nil, main_employer_siret: '123')
        hr.reload.update(consent_level: 3)

        expect(subject).to match_array([
          "Situation professionnelle Nom de l'employeur (si vous en avez plusieurs, employeur principal)",
        ])
      end 

      it 'main_employer_siret not blank but CHO' do 
        hr.get_applicant_used_for_siret.update(main_employer_name: nil, main_employer_siret: '123', work_contract_type: 'CHO')
        hr.reload.update(consent_level: 3)

        expect(subject).to match_array([])
      end 
    end 

    context 'consent_level' do 
      it 'when entity_id no consent level' do 
        hr.get_applicant_used_for_siret.update(main_employer_siret: entity.siret)

        expect(subject).to match_array([
          "Niveau de consentement Niveau de consentement non reseigné",
        ])
      end 

      it 'when entity_id with consent level' do 
        hr.get_applicant_used_for_siret.update(main_employer_siret: entity.siret)
        hr.reload.update(consent_level: 3)
        expect(subject).to match_array([])
      end 

      it 'when consent_level 1' do 
        hr.get_applicant_used_for_siret.update(main_employer_siret: entity.siret)
        hr.reload.update(consent_level: 1)
        expect(subject).to match_array([])
      end 

      it 'when no entity but CDI' do 
        hr.get_applicant_used_for_siret.update(main_employer_siret: "test", work_contract_type: 'CDI')
        
        expect(subject).to match_array([
          "Niveau de consentement Niveau de consentement non reseigné",
        ])
      end 
    end

    context 'main_employer_siret' do 
      %w[
        CDI
        CDIP
      ].each do |contract_type| 

        it "no error when siret and #{contract_type}" do 
          hr.get_applicant_used_for_siret.update(main_employer_siret: entity.siret, work_contract_type: contract_type)
          hr.reload.update(consent_level: 2)
          expect(subject).to match_array([])
        end 
  
        it "no error when no siret and #{contract_type}" do 
          hr.get_applicant_used_for_siret.update(main_employer_siret: nil, work_contract_type: contract_type)
          hr.reload.update(consent_level: 2)

          expect(subject).to match_array([
            "Situation professionnelle Demandeur Siret",
          ])
        end 
      end 

      %w[
        CHO
        AGP
        APP
        ASS
        AGH
        ACT
        IND
        RET
        AUT
      ].each do |contract_type| 
        it "no error when no siret and #{contract_type}" do
          hr.get_applicant_used_for_siret.update(main_employer_siret: nil, work_contract_type: contract_type)
          hr.reload.update(consent_level: 2)

          expect(subject).to match_array([])
        end 
      end 

    end
  end

  describe 'update_entity_relationship' do 
    let!(:entity1) {create(:entity, siret: '10000000000000', siren: '100000000')}
    let!(:entity11) {create(:entity, siret: '10000000000001', siren: '100000000', parent_id: entity1.id)}
    let!(:entity12) {create(:entity, siret: '10000000000002', siren: '100000000', parent_id: entity1.id)}
    let!(:entity2) {create(:entity, siret: '20000000000000', siren: '200000000')}  
    let!(:entity21) {create(:entity, siret: '20000000000001', siren: '200000000', parent_id: entity2.id)}
    let!(:entity3) {create(:entity, siret: '30000000000000', siren: '300000000')}  
    let!(:entity31) {create(:entity, siret: '30000000000001', siren: '300000001', parent_id: entity3.id)}  
    let!(:entity32) {create(:entity, siret: '30000000000002', siren: '300000001', parent_id: entity3.id)} 

    let!(:entity_user){ create(:user)}
    let!(:applicant){ create(:applicant, main_employer_siret: entity1.siret)}
    let!(:coapp){ create(:applicant)}
    let!(:entity_code){ create(:entity_code, entity_id: entity2.id, code: 'ENTITY2' )}
    let!(:hr) { create(:housing_request, main_applicant: applicant, is_active: true, user: user, coapplicants: [{id: coapp.id.to_s, rel_type: 'C'}], consent_level: 2)}

    it 'default' do 
      expect(hr.reload.entity_id).to eq(entity1.id)
      expect(hr.reload.consent_level).to eq(2)
    end 

    it 'when last_entity_code' do 
      hr.update(last_entity_code: entity_code.code)
      expect(hr.reload.entity_id).to eq(entity2.id)
      expect(hr.reload.consent_level).to eq(nil)
      hr.update(last_entity_code: nil)
      expect(hr.reload.entity_id).to eq(entity1.id)
    end 

    it 'when last_entity_code changed from consent level 1' do
      hr.update(consent_level: 1)
      expect(hr.reload.entity_id).to eq(nil)
      expect(hr.reload.consent_level).to eq(1)
      hr.update(last_entity_code: entity_code.code)
      expect(hr.reload.entity_id).to eq(entity2.id)
      expect(hr.reload.consent_level).to eq(nil)
    end 

    it 'when main_employer_siret change' do 
      applicant.update(main_employer_siret: entity3.siret)
      expect(hr.reload.entity_id).to eq(entity3.id)
      expect(hr.reload.consent_level).to eq(nil)
    end 

    it 'when coapplicant main_employer_siret change' do 
      hr.update(applicant_used_for_siret: coapp.id)
      coapp.update(main_employer_siret: entity3.siret)
      expect(hr.reload.entity_id).to eq(entity3.id)
      expect(hr.reload.consent_level).to eq(nil)
      expect(hr.applicant_used_for_siret_firstname).to eq(coapp.first_name)
      coapp.first_name = "dudule"
      coapp.save
      expect(hr.reload.applicant_used_for_siret_firstname).to eq("dudule")
    end 

    it 'when siret not 14 length' do
      applicant.update(main_employer_siret: '2000000000000')
      expect(hr.reload.entity_id).to eq(nil)
      expect(hr.reload.consent_level).to eq(nil)
    end 

    it 'when siret not 14 length' do
      applicant.update(main_employer_siret: '2000000000000')
      expect(hr.reload.entity_id).to eq(nil)
      expect(hr.reload.consent_level).to eq(nil)
    end 

    it 'when no siren found' do
      applicant.update(main_employer_siret: '30000000100000')
      expect(hr.reload.entity_id).to eq(nil)
      expect(hr.reload.consent_level).to eq(nil)
    end 

    it 'when one siren entities' do 
      entity2.update(siret: '12345678901234', siren: '123456789')
      applicant.update(main_employer_siret: "#{entity2.siren}43210")
      expect(hr.reload.entity_id).to eq(entity2.id)
    end 

    it 'when only one root' do
      applicant.update(main_employer_siret: '20000000000005')
      expect(hr.reload.entity_id).to eq(entity2.id)
      expect(hr.reload.consent_level).to eq(nil)
    end 

    it 'when only one master and one root_id' do
      entity11.update(is_master: true)
      applicant.update(main_employer_siret: '10000000000005')
      expect(hr.reload.entity_id).to eq(entity1.id)
    end 

    it 'when only one master and multiple root_id' do
      entity11.update(is_master: true)
      entity2.update(siret: '10000000000005', siren: '100000000')
      applicant.update(main_employer_siret: '10000000000006')
      expect(hr.reload.entity_id).to eq(entity11.id)
    end 

    it 'when multiple master and multiple root_id' do
      entity11.update(is_master: true)
      entity12.update(is_master: true)
      entity2.update(siret: '10000000000005', siren: '100000000')
      applicant.update(main_employer_siret: '10000000000006')
      expect(hr.reload.entity_id).to eq(nil)
    end 

    it 'when CHO consent_level cleared' do 
      expect { applicant.update(work_contract_type: 'APP') }.to change { hr.reload.consent_level }.from(2).to(nil)
    end 

    describe 'clear last_entity_code' do
      let!(:entity_code3){ create(:entity_code, entity_id: entity3.id, code: 'ENTITY3' )}

      before do 
        hr.update(last_entity_code: entity_code.code)
        coapp.update(main_employer_siret: entity1.siret)
      end 

      it 'cleared applicant_used_for_siret change' do
        hr.update(applicant_used_for_siret: coapp.id)
        expect(hr.reload.last_entity_code).to eq(nil)
        expect(hr.reload.entity_id).to eq(entity1.id)
      end 
  
      it 'not cleared when applicant_used_for_siret and last_entity_code change' do
        hr.update(applicant_used_for_siret: coapp.id, last_entity_code: entity_code3.code)
        expect(hr.reload.last_entity_code).to eq('ENTITY3')
        expect(hr.reload.entity_id).to eq(entity3.id)
      end 
    end 
  end
  
  describe 'clear_attributes' do 
    let!(:applicant) {create(:applicant, main_employer_name: 'Test')}
    let!(:hr) {create(:housing_request, main_applicant: applicant, is_active: true)}

    it 'works' do 
      hr.update(is_active: false)
      expect(applicant.reload.main_employer_name).to eq(nil)
    end 
  end 

  describe 'update_applicant_used_for_siret_info' do 
    let!(:app1) { create(:applicant, first_name: 'One', last_name: 'Doe', email: 'one@test.com')}
    let!(:app2) { create(:applicant, first_name: 'Two', last_name: 'Doe', email: 'two@test.com')}
    let!(:hr) {create(:housing_request, is_active: true, main_applicant: app1, coapplicants: [{id: app2.id.to_s, rel_type: 'C'}])}

    it 'default' do 
      expect(hr.applicant_used_for_siret_firstname).to eq('One')
      expect(hr.applicant_used_for_siret_lastname).to eq('Doe')
      expect(hr.applicant_used_for_siret_fullname).to eq('One Doe')
      expect(hr.applicant_used_for_siret_email).to eq('one@test.com')
      expect(hr.applicant_used_for_siret_birth_date).to eq(app1.birth_date)
    end 

    it 'when main_applicant email change' do 
      expect {
        app1.update(email: 'onebis@test.com')
      }.to change { hr.reload.applicant_used_for_siret_email }.from('one@test.com').to('onebis@test.com')
    end 

    it 'when applicant_used_for_siret change' do 
      hr.update(applicant_used_for_siret: app2.id)
      hr.reload
      expect(hr.applicant_used_for_siret_firstname).to eq('Two')
      expect(hr.applicant_used_for_siret_lastname).to eq('Doe')
      expect(hr.applicant_used_for_siret_fullname).to eq('Two Doe')
      expect(hr.applicant_used_for_siret_email).to eq('two@test.com')
      expect(hr.applicant_used_for_siret_birth_date).to eq(app2.birth_date)
    end 

    it 'when applicant_used_for_siret and email missing' do 
      hr.update(applicant_used_for_siret: app2.id)
      app2.update(email: nil)
      hr.reload
      expect(hr.applicant_used_for_siret_firstname).to eq('Two')
      expect(hr.applicant_used_for_siret_lastname).to eq('Doe')
      expect(hr.applicant_used_for_siret_fullname).to eq('Two Doe')
      expect(hr.applicant_used_for_siret_email).to eq('one@test.com')
      expect(hr.applicant_used_for_siret_birth_date).to eq(app2.birth_date)
    end 

    it 'when applicant_used_for_siret and email missing' do 
      app2.update(email: "twobis@test.com")
      hr.reload
      expect(hr.applicant_used_for_siret_firstname).to eq('One')
      expect(hr.applicant_used_for_siret_lastname).to eq('Doe')
      expect(hr.applicant_used_for_siret_fullname).to eq('One Doe')
      expect(hr.applicant_used_for_siret_email).to eq('one@test.com')
      expect(hr.applicant_used_for_siret_birth_date).to eq(app1.birth_date)
    end 
  end 

  describe 'entity_status workflow' do 
    let!(:hr) { create(:complete_request, user: user)}
    let!(:offer) { create(:housing_offer, district: 'Paris', postal_code: '75000')}
    let!(:entity) { create(:entity, siret: '12345678901234')}

    it 'no entity_id' do 
      expect(hr.reload.entity_status).to eq(nil)
      
      ha = Hermes::HousingApplication.create(status: 'EN ATTENTE', housing_offer: offer, housing_request: hr )
      expect(hr.reload.entity_status).to eq(nil)

      ha.update!(status: 'TRANSMIS BAILLEUR')
      expect(hr.reload.entity_status).to eq(nil)

      ha.update!(status: 'NON RETENUE')
      expect(hr.reload.entity_status).to eq(nil)

      ha.update!(status: 'LOGE')
      expect(hr.reload.entity_status).to eq(nil)
      expect(hr.request_status).to eq('SATISFAITE')
      expect(hr.attributed_housing_district).to eq('Paris')
      expect(hr.attributed_housing_district_postal_code).to eq('75000')
    end

    it 'with cancelled hr' do 
      hr.main_applicant.update(main_employer_siret: entity.siret)
      hr.reload.update(form_status: 'cancelled')
      expect(hr.reload.request_status).to eq('ANNULEE')
      expect(hr.reload.entity_status).to eq('ANNULE')
    end 

    it 'with entity_id' do 
      hr.main_applicant.update(main_employer_siret: entity.siret)
      hr.reload.update(consent_level: 2)
      expect(hr.reload.entity_status).to eq('EN COURS')

      entity.update(auto_request_validation: false)
      hr.update(entity_validated: false)
      expect(hr.reload.entity_status).to eq('A VALIDER')

      hr.update(entity_validated: true)
      expect(hr.reload.entity_status).to eq('EN COURS')
    end 

    it 'with REFUSE' do 
      entity.update(auto_request_validation: false )
      hr.main_applicant.update(main_employer_siret: entity.siret)
      hr.reload.update(consent_level: 2)
      expect(hr.reload.entity_status).to eq('A VALIDER')

      hr.update(entity_refused: true)
      expect(hr.reload.entity_status).to eq('REFUSE')

      ha = Hermes::HousingApplication.create(status: 'EN ATTENTE', housing_offer: offer, housing_request: hr )
      expect(hr.reload.entity_status).to eq('REFUSE')

      ha.update!(status: 'TRANSMIS BAILLEUR')
      expect(hr.reload.entity_status).to eq('REFUSE')

      ha.update!(status: 'NON RETENUE')
      expect(hr.reload.entity_status).to eq('REFUSE')

      hr.set(active_application_id: ha.id)
      ha.update!(status: 'LOGE')
      expect(hr.reload.entity_status).to eq('REFUSE')
      expect(hr.reload.request_status).to eq('SATISFAITE')
    end 

    it 'with A VALIDER' do 
      entity.update(auto_request_validation: false)
      hr.main_applicant.update(main_employer_siret: entity.siret)
      hr.reload.update(consent_level: 2)
      expect(hr.reload.entity_status).to eq('A VALIDER')

      ha = Hermes::HousingApplication.create(status: 'EN ATTENTE', housing_offer: offer, housing_request: hr )
      expect(hr.reload.entity_status).to eq('A VALIDER')

      ha.update!(status: 'TRANSMIS BAILLEUR')
      expect(hr.reload.entity_status).to eq('A VALIDER')

      ha.update!(status: 'NON RETENUE')
      expect(hr.reload.entity_status).to eq('A VALIDER')

      hr.set(active_application_id: ha.id)
      ha.update!(status: 'LOGE')
      expect(hr.reload.entity_status).to eq('A VALIDER')
      expect(hr.reload.request_status).to eq('SATISFAITE')
    end 

    it 'with EN COURS' do 
      hr.main_applicant.update(main_employer_siret: entity.siret)
      hr.update(consent_level: 2)
      expect(hr.reload.entity_status).to eq('EN COURS')

      ha = Hermes::HousingApplication.create(status: 'EN ATTENTE', housing_offer: offer, housing_request: hr )
      expect(hr.reload.entity_status).to eq('EN COURS')

      ha.update!(status: 'TRANSMIS BAILLEUR')
      expect(hr.reload.entity_status).to eq('DOSSIER PRESENTE')

      ha.update!(status: 'NON RETENUE')
      expect(hr.reload.entity_status).to eq('EN COURS')

      hr.set(active_application_id: ha.id)
      ha.update!(status: 'RETENUE EN CAL')
      expect(hr.reload.entity_status).to eq('DOSSIER ACCEPTE')

      ha.update!(status: 'LOGE')
      expect(hr.reload.entity_status).to eq('SALARIE LOGE')
      expect(hr.reload.request_status).to eq('SATISFAITE')
    end 
  end 

  describe 'entity_refused' do 
    let!(:entity) { create(:entity, siret: '12345678901234')}
    let!(:app) { create(:applicant, main_employer_siret: entity.siret)}
    let!(:hr) {create(:housing_request, main_applicant: app)}

    it 'default' do 
      expect(hr.reload.entity_refused).to eq(false)
      expect(hr.reload.entity_status).not_to eq('REFUSE')
    end 

    it 'when true' do 
      hr.update(entity_refused: true)
      expect(hr.reload.entity_status).to eq('REFUSE')
    end 

    it 'force entity_valided' do
      hr.set(entity_validated: true)
      hr.update(entity_refused: true)
      expect(hr.reload.entity_validated).to eq(false)
    end 
  end 

  describe 'only_public eligibility' do 
    let!(:hr) { create(:complete_request)}
    let!(:app) { hr.main_applicant }
    subject { hr.reload.eligibility_query }

    it 'default' do 
      expect(subject).not_to eq({is_public: true})
    end 

    it 'when is inactive' do 
      hr.update(is_active: false)
      expect(subject).to eq({is_public: true})
    end 

    it 'not when reference_tax_revenues 0' do 
      hr.update(reference_tax_revenues: 0)
      expect(subject).not_to eq({is_public: true})
    end 

    it 'when reference_tax_revenues nil' do 
      hr.update(reference_tax_revenues: nil)
      expect(subject).to eq({is_public: true})
    end 

    it 'not when monthly_revenues 0' do 
      hr.update(monthly_revenues: 0)
      expect(subject).not_to eq({is_public: true})
    end 

    it 'when monthly_revenues nil' do 
      hr.update(monthly_revenues: nil)
      expect(subject).to eq({is_public: true})
    end 

    it 'all_eligibility_query' do 
      hr.update(monthly_revenues: nil)
      expect(hr.all_eligibility_query).to eq({is_public: true})
    end 
  end

  describe 'form_status' do 
    let!(:hr) { create(:complete_request)}

    it 'works by default' do 
      expect(hr.reload.form_status).to eq('complete')
      expect(hr.reload.request_status).to eq('COMPLETE')
    end 

    it 'when cancelling request' do 
      hr.reload.update(form_status: 'cancelled')
      expect(hr.reload.form_status).to eq('cancelled')
      expect(hr.reload.request_status).to eq('ANNULEE')
      hr.reload.update(seeked_departments: [])
      expect(hr.reload.form_status).to eq('cancelled')
      expect(hr.reload.request_status).to eq('ANNULEE')
    end 

    it 'when missing truly_seeked_departments' do 
      hr.reload.update(seeked_departments: [], seeked_district_insee_codes: [])
      expect(hr.reload.form_status).to eq('invalid')
      expect(hr.reload.request_status).to eq('INCOMPLETE')
      hr.reload.update(form_status: 'cancelled')
      expect(hr.reload.form_status).to eq('cancelled')
      expect(hr.reload.request_status).to eq('ANNULEE')
    end 

    it 'when form_front_validation_errors' do 
      hr.main_applicant.reload.update(main_employer_name: nil)
      expect(hr.reload.form_status).to eq('invalid')
      expect(hr.reload.request_status).to eq('INCOMPLETE')
      hr.reload.update(form_status: 'cancelled')
      expect(hr.reload.form_status).to eq('cancelled')
      expect(hr.reload.request_status).to eq('ANNULEE')
    end 
  end 
  
  describe 'applicants_attached_codes eligibility' do 
    let!(:entity) { create(:entity, siret: '12345678901234') }
    let!(:applicant) {create(:applicant, :with_eligible_offers, offers_count: 3)}
    let!(:hr) {applicant.first_active_request }


    it 'works' do 
      expect(hr.applicants_attached_codes).to eq(['34100'])
      expect(hr.reload.eligibility_query[:$or].last).to eq({housing_channel_keys: 'township_attachment', insee_code: { '$in': ['34100']}.stringify_keys}.stringify_keys)

      expect(hr.reload.eligibles_offers.count).to eq(3)

      hr.eligibles_offers.each_with_index do |o, i|
        case i
        when 0
          o.set(housing_channel_keys: ['dc', 'township_attachment'], insee_code: '34100')
        when 1
          o.set(housing_channel_keys: ['dc'], insee_code: '34100')
        when 2
          o.set(housing_channel_keys: ['dc', 'township_attachment'], insee_code: '39000')
        end
      end 
      expect(hr.reload.eligibles_offers.count).to eq(2)

      applicant.update(workplace_insee_code: '95000')
      expect(hr.reload.applicants_attached_codes).to eq(['95000'])
      expect(hr.reload.eligibility_query[:$or].last).to eq({housing_channel_keys: 'township_attachment', insee_code: { '$in': ['95000']}.stringify_keys}.stringify_keys)
      expect(hr.reload.eligibles_offers.count).to eq(1)

      # check reserved offer
      applicant.update(main_employer_siret: entity.siret)
      expect(hr.reload.entity_id).to eq(entity.id)
      expect(hr.reload.entity_validated).to eq(true)

      expect(hr.reload.eligibles_offers.count).to eq(1)
      hr.reload.eligibles_offers.first.set(reserved: true, active_reservations: [entity.id.to_s])
      expect(hr.reload.eligibles_offers.count).to eq(1)

    end 

  end 

  describe 'eligibility_query types' do 
    let!(:entity) { create(:entity)}
    let!(:bt1) { create(:bordering_town, insee_code: '34200', bordering_towns: ['34210'])}
    let!(:bt2) { create(:bordering_town, insee_code: '34100', bordering_towns: ['34110', '34111'])}
    let!(:hr) { create(:complete_request, seeked_district_insee_codes: ['34200', '34100'])}
    let!(:default_offer_attributes) { {typology: 'T1', ceil_category_1: 25000, floor_category_1: 0}}
    let!(:offer_in_department_1) { create(:published_offer, default_offer_attributes.merge(insee_code: '34000')) }
    let!(:offer_seeked) { create(:published_offer, default_offer_attributes.merge(insee_code: '34200')) }
    let!(:offer_bordering) { create(:published_offer, default_offer_attributes.merge(insee_code: '34110')) }
    let!(:offer_in_departement_2) { create(:published_offer, default_offer_attributes.merge(insee_code: '34300')) }
    let!(:offer_out_of_departement) { create(:published_offer, default_offer_attributes.merge(insee_code: '99000')) }

    it 'works' do
      # department
      expect(hr.eligibility_query[:department]["$in"]).to eq(['34'])
      expect(hr.eligibles_offers.count).to eq(4)

      # seeked
      expect(hr.eligibility_query('seeked')[:insee_code]["$in"].sort).to eq(['34200', '34100'].sort)
      expect(hr.eligibles_offers('seeked').count).to eq(1)
      expect(hr.is_eligible_offer(offer_seeked, 'seeked')).to eq(true)

      # bordering
      expect(hr.eligibility_query('bordering')[:insee_code]["$in"].sort).to eq(['34210', '34110', '34111'].sort)
      expect(hr.eligibles_offers('bordering').count).to eq(1)
      expect(hr.is_eligible_offer(offer_bordering, 'bordering')).to eq(true)

      # seeked_bordering
      expect(hr.eligibility_query('seeked_bordering')[:insee_code]["$in"].sort).to eq(['34200', '34100', '34210', '34110', '34111'].sort)
      expect(hr.eligibles_offers('seeked_bordering').count).to eq(2)
      expect(hr.is_eligible_offer(offer_seeked, 'seeked_bordering')).to eq(true)
      expect(hr.is_eligible_offer(offer_bordering, 'seeked_bordering')).to eq(true)
    
      # other_from_department
      expect(hr.eligibility_query('other_from_department')[:insee_code]["$nin"].sort).to eq(['34200', '34100', '34210', '34110', '34111'].sort)
      expect(hr.eligibility_query('other_from_department')[:department]["$in"]).to eq(['34'])
      expect(hr.eligibles_offers('other_from_department').count).to eq(2)
      expect(hr.is_eligible_offer(offer_in_department_1, 'other_from_department')).to eq(true)
      expect(hr.is_eligible_offer(offer_in_departement_2, 'other_from_department')).to eq(true)

      # all_eligibility_query
      hr.set(entity_id: entity.id, entity_validated: true)
      hr.update_eligibility_query
      # expect(hr.all_eligibility_query[:$or].first).to eq({"housing_channel_keys" => { "$ne" => 'township_attachment'}})
      expect(hr.all_eligibility_query[:$and].first["$or"].first).to eq({"reserved" => true, "active_reservations" => hr.entity_id.to_s})
      expect(hr.all_eligibility_query[:$and].last[:$and].first[:$or].first).to eq({department: { "$in" => [ "34"]}})

      expect(Hermes::HousingOffer.where(hr.all_eligibility_query).count).to eq(4)
    end 

    it 'when missing data' do 
      hr.update(truly_seeked_departments: [], seeked_departments: [], seeked_district_insee_codes: [], seeked_district_postal_codes: [])
      hr.reload
      expect(hr.eligibles_offers('department').count).to eq(0)
      expect(hr.eligibles_offers('seeked').count).to eq(0)
      expect(hr.eligibles_offers('seeked_bordering').count).to eq(0)
      expect(hr.eligibles_offers('other_from_department').count).to eq(0)
      expect(Hermes::HousingOffer.where(hr.all_eligibility_query).count).to eq(0)
    end 

  end 

  describe 'BOROUGHS' do 
    let!(:hr) { create(:complete_request)}

    it 'works' do 
      hr.update(seeked_localisation_list: [{code: '75056'}, {code: '10000'}])
      expect(hr.reload.seeked_district_insee_codes.length).to eq(22)
    end 
  end

  describe 'indexes eligibility' do 
    let!(:hr) { create(:complete_request)}
    
    it 'works' do
      Hermes::HousingOffer.remove_indexes
      Hermes::HousingOffer.create_indexes
  
      expect(hr.eligibles_offers.explain.to_s).not_to include("COL")
      expect(hr.eligibles_offers.explain['queryPlanner']['winningPlan']['inputStage']['indexName']).to eq('eligibility_base')
    end 
  end 

  describe 'bordering_insee_codes' do 
    let!(:bt1) { create(:bordering_town, insee_code: '20000', bordering_towns: ['20001', '20000'])}
    let!(:bt2) { create(:bordering_town, insee_code: '21000', bordering_towns: ['21001', '22000'])}
    let!(:bt3) { create(:bordering_town, insee_code: '75000', bordering_towns: ['75001', '77000', '29000'])}
    let!(:bt4) { create(:bordering_town, insee_code: '97000', bordering_towns: ['97001', '97201'])}

    let!(:hr) { create(:active_request)}


    it 'same departement' do 
      hr.update(seeked_localisation_list: [{code: '20000'}])
      expect(hr.reload.bordering_insee_codes).to eq(['20001'])
    end 

    it 'not same departement' do 
      hr.update(seeked_localisation_list: [{code: '21000'}])
      expect(hr.reload.bordering_insee_codes).to eq(['21001'])
    end 

    it 'when in idf' do 
      hr.update(seeked_localisation_list: [{code: '75000'}])
      expect(hr.reload.bordering_insee_codes).to eq(['75001', '77000'])
    end 

    it 'when > 95' do 
      hr.update(seeked_localisation_list: [{code: '97000'}])
      expect(hr.reload.bordering_insee_codes).to eq(['97001'])
    end 

    it 'multiple departement' do 
      hr.update(seeked_localisation_list: [{code: '20000'}, {code: '21000'}, {code: '75000'}])
      expect(hr.reload.bordering_insee_codes).to eq(['20001', '21001', '75001', '77000'])
    end 


  end 

  describe 'flux36' do 
    let!(:hr) {create(:complete_request)}

    it 'send when all params' do 
      expect do
        hr.update(als_id: '1', entity_id: '1', first_flux_30_sended: true )
      end.to change { enqueued_jobs.select{|a| a[:job] == ExecuteFlux36TaskJob }.length }.from(0).to(1)
    end

     [:als_id, :entity_id, :first_flux_30_sended].each do |att|
      it "not send when missing #{att}" do 
        data = {als_id: '1', entity_id: '1', first_flux_30_sended: true }
        hr.update(data.except(att))
        expect( enqueued_jobs.select{|a| a[:job] == ExecuteFlux36TaskJob }.length).to eq(0)
      end 
     end 

  end 

  describe 'cerfa v5' do 
    let!(:hr) {create(:active_request)}
    let!(:nud) {create(:nud, housing_request_id: hr.id, unique_number: 'cerfa_v5_new_champs' )}

    context 'new fields' do 
      it "default" do 
        expect(hr.reload.cerfa_version).to eq("3")
        expect(hr.reload.assistant_phone_number).to be_nil
        expect(hr.reload.assistant_notifications).to be_nil
        expect(hr.reload.attributed_housing_surface).to be_nil
        expect(hr.reload.nb_to_accomodate).to be_nil
      end 
  
      it 'updated' do 
        hr.reload.update_request_from_sne(user)
        expect(hr.reload.cerfa_version).to eq("5")
        expect(hr.reload.assistant_phone_number).to eq('0102030405')
        expect(hr.reload.assistant_notifications).to eq(true)
        expect(hr.reload.nb_to_accomodate).to eq(12)
        expect(hr.reload.attributed_housing_surface).to be_nil
      end
    end   

    context 'new work_contract_type' do 
      let!(:hr) {create(:complete_request)}
      let(:entity) { create(:entity, siret: '12345678901234') }

      before do 
        hr.main_applicant.update(main_employer_siret: entity.siret)
        hr.reload.update(entity_validated: true, priority_points_indexes: [100])
      end 

      it 'no entity_id for CHO' do 
        expect {
          hr.main_applicant.update(work_contract_type: 'CHO')
        }.to change {hr.reload.entity_id}.from(entity.id).to(nil)
      end 

      ["CDIP", "CDI", "AGP", "APP", "ASS", "AGH", "ACT", "IND", "RET", "AUT"].each do |type|
        it "had score only when work_contract_type #{type}" do 
          hr.main_applicant.update(work_contract_type: type)
          hr.reload.update(consent_level: 2)

          expect(hr.reload.score_details['entity_validated_points']).to eq(240)
          expect(hr.reload.score_details['priority_points']).to eq(100)
        end 
      end 

      ["CHO", "", nil].each do |type|
        it "had not score when work_contract_type #{type}" do 
          hr.main_applicant.update(work_contract_type: type)
          hr.reload.update(consent_level: 2)

          expect(hr.reload.score_details['entity_validated_points']).to eq(0)
          expect(hr.reload.score_details['priority_points']).to eq(0)
        end 
      end 


    end 

    context 'new motivations' do 
      let!(:hr) {create(:complete_request)}

      it 'add new motivations' do 
        hr.reload.update_request_from_sne(user)
        expect(hr.reload.motivations).to include('29')
        expect(hr.reload.motivations).to include('30')
        expect(hr.reload.motivations).to include('32')
        expect(hr.reload.motivations.count('31')).to eq(2)
      end 

      ['29', '30', '06'].each do |motivation|
        it "add violence points for #{motivation}" do 
          expect {
            hr.update(motivations: [motivation])
          }.to change {hr.reload.score_details['violence_points']}.from(0).to(40)
        end 
      end 
      
      it 'when mutliple violence motivations' do
        expect {
          hr.update(motivations: ['29', '30', '06'])
        }.to change {hr.reload.score_details['violence_points']}.from(0).to(40)
      end 

      it 'when motivations nil' do 
        hr.update(motivations: nil)
        expect(hr.reload.score_details['violence_points']).to eq(0)
      end 
    end 
  end 

  describe 'not calculate motivations' do 
    let!(:hr) {create(:active_request)}
    let!(:nud) {create(:nud, housing_request_id: hr.id, unique_number: 'not_calculate_motivations' )}

    it 'works' do 
      hr.reload.update_request_from_sne(user)
      expect(hr.reload.motivations).not_to include("26")
      expect(hr.reload.motivations).not_to include("27")
    end
  end 
  
  describe 'only_one_active_request_per_user' do
    let!(:hr) { create(:active_request, user: user) }
    let!(:app) { create(:applicant) }

    it 'throw an error' do 
      hr2 = Hermes::HousingRequest.new(user: user, is_active: true, main_applicant: app)
      expect(hr2.valid?).to eq(false)
      expect(hr2.errors.messages[:user_id]).to eq(['Vous avez déja une demande active'])
    end 

  end 
end
